#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Standard plug-in to make dill module work for compiled stuff."""

from nuitka.options.Options import shallMakeModule
from nuitka.plugins.PluginBase import (
    NuitkaNamespaceDetectorPluginBase,
    NuitkaPluginBase,
)


class NuitkaPluginDillWorkarounds(NuitkaPluginBase):
    """This is to make dill/ray/cloudpickle modules work with compiled methods."""

    plugin_name = "dill-compat"
    plugin_desc = "Required by 'dill' and 'cloudpickle' packages."
    plugin_category = "package-support"

    pickle_package_names = ("dill", "cloudpickle", "ray")

    @staticmethod
    def isAlwaysEnabled():
        return False

    def __init__(self, pickle_supported_modules):
        self.pickle_supported_modules = pickle_supported_modules or ["all"]

        self.pickle_selected_packages = [
            self._getPickleSupportPackageName(pickle_package_name)
            for pickle_package_name in self.pickle_package_names
            if self.shallIncludePickleSupportModule(pickle_package_name)
        ]

    def shallIncludePickleSupportModule(self, name):
        return (
            name in self.pickle_supported_modules
            or "all" in self.pickle_supported_modules
        )

    @staticmethod
    def _getPickleSupportPackageName(name):
        if name == "ray":
            return "ray.cloudpickle"

        return name

    def _getPostLoadCode(self, name):
        if name == "ray.cloudpickle":
            result = self._getPostLoadCode("cloudpickle")

            assert "import cloudpickle" in result
            result = result.replace(
                "import cloudpickle", "import ray.cloudpickle as cloudpickle"
            )
            return result

        return self.getPluginDataFileContents("%s-postLoad.py" % name)

    @classmethod
    def addPluginCommandLineOptions(cls, group):
        group.add_option(
            "--include-pickle-support-module",
            action="append",
            dest="pickle_supported_modules",
            choices=("all",) + cls.pickle_package_names,
            default=[],
            help="""\
Include support for these modules to pickle nested compiled functions. You
can use "all" which is the default, but esp. in module mode, just might
want to limit yourself to not create unnecessary run-time usages. For
standalone mode, you can leave it at the default, at it will detect
the usage.""",
        )

    def createPostModuleLoadCode(self, module):
        full_name = module.getFullName()

        if not shallMakeModule():
            for candidate in self.pickle_selected_packages:
                if full_name == candidate:
                    return (
                        self._getPostLoadCode(candidate),
                        """\
Extending "%s" for compiled types to be pickle-able as well."""
                        % candidate,
                    )
        elif module.isTopModule():
            return (
                """\
import sys
sys.modules[__compiled__.main]._create_compiled_function = \
    sys.modules[__name__.replace("-postLoad", "-preLoad")]._create_compiled_function
sys.modules[__compiled__.main]._create_compiled_function.__module__ = \
    __compiled__.main
""",
                """
Extending for compiled types to be pickle-able as well.""",
            )

    def createPreModuleLoadCode(self, module):
        if shallMakeModule() and module.isTopModule():
            for candidate in self.pickle_selected_packages:
                if self.shallIncludePickleSupportModule(candidate):
                    yield (
                        self._getPostLoadCode(candidate),
                        """\
Extending "%s" for compiled types to be pickle-able as well.""",
                    )

    def onModuleEncounter(
        self, using_module_name, module_name, module_filename, module_kind
    ):
        if module_name.hasOneOfNamespaces(*self.pickle_selected_packages):
            return True, "Needed to handle %s" % module_name.getTopLevelPackageName()

    @staticmethod
    def getPreprocessorSymbols():
        return {"_NUITKA_PLUGIN_DILL_ENABLED": "1"}

    def getExtraCodeFiles(self):
        return {"DillPlugin.c": self.getPluginDataFileContents("DillPlugin.c")}


class NuitkaPluginDetectorDillWorkarounds(NuitkaNamespaceDetectorPluginBase):
    """Only used if plugin is NOT activated.

    Notes:
        We are given the chance to issue a warning if we think we may be required.
    """

    detector_for = NuitkaPluginDillWorkarounds
    detector_namespaces = NuitkaPluginDillWorkarounds.pickle_package_names


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.gnu.org/licenses/agpl.txt
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



# MIT License
#
# Copyright The SCons Foundation
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
# KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""
Windows registry functions for Microsoft Visual C/C++.
"""

import os

from SCons.Util import (
    HKEY_LOCAL_MACHINE,
    HKEY_CURRENT_USER,
    RegGetValue,
)

from .. common import (
    debug,
)

from . import Util

from . import Dispatcher
Dispatcher.register_modulename(__name__)


# A null-terminated string that contains unexpanded references to environment variables.
REG_EXPAND_SZ = 2

def read_value(hkey, subkey_valname, expand=True):
    try:
        rval_t = RegGetValue(hkey, subkey_valname)
    except OSError:
        debug('OSError: hkey=%s, subkey=%s', repr(hkey), repr(subkey_valname))
        return None
    rval, regtype = rval_t
    if regtype == REG_EXPAND_SZ and expand:
        rval = os.path.expandvars(rval)
    debug('hkey=%s, subkey=%s, rval=%s', repr(hkey), repr(subkey_valname), repr(rval))
    return rval

def registry_query_path(key, val, suffix, expand=True):
    extval = val + '\\' + suffix if suffix else val
    qpath = read_value(key, extval, expand=expand)
    if qpath and os.path.exists(qpath):
        qpath = Util.normalize_path(qpath)
    else:
        qpath = None
    return (qpath, key, val, extval)

REG_SOFTWARE_MICROSOFT = [
    (HKEY_LOCAL_MACHINE, r'Software\Wow6432Node\Microsoft'),
    (HKEY_CURRENT_USER,  r'Software\Wow6432Node\Microsoft'), # SDK queries
    (HKEY_LOCAL_MACHINE, r'Software\Microsoft'),
    (HKEY_CURRENT_USER,  r'Software\Microsoft'),
]

def microsoft_query_paths(suffix, usrval=None, expand=True):
    paths = []
    records = []
    for key, val in REG_SOFTWARE_MICROSOFT:
        extval = val + '\\' + suffix if suffix else val
        qpath = read_value(key, extval, expand=expand)
        if qpath and os.path.exists(qpath):
            qpath = Util.normalize_path(qpath)
            if qpath not in paths:
                paths.append(qpath)
                records.append((qpath, key, val, extval, usrval))
    return records

def microsoft_query_keys(suffix, usrval=None, expand=True):
    records = []
    for key, val in REG_SOFTWARE_MICROSOFT:
        extval = val + '\\' + suffix if suffix else val
        rval = read_value(key, extval, expand=expand)
        if rval:
            records.append((rval, key, val, extval, usrval))
    return records

def microsoft_sdks(version):
    return '\\'.join([r'Microsoft SDKs\Windows', 'v' + version, r'InstallationFolder'])

def sdk_query_paths(version):
    q = microsoft_sdks(version)
    return microsoft_query_paths(q)

def windows_kits(version):
    return r'Windows Kits\Installed Roots\KitsRoot' + version

def windows_kit_query_paths(version):
    q = windows_kits(version)
    return microsoft_query_paths(q)

def vstudio_sxs_vs7(version):
    return '\\'.join([r'VisualStudio\SxS\VS7', version])

def vstudio_sxs_vc7(version):
    return '\\'.join([r'VisualStudio\SxS\VC7', version])

def devdiv_vs_servicing_component(version, component):
    return '\\'.join([r'DevDiv\VS\Servicing', version, component, 'Install'])




# MIT License
#
# Copyright The SCons Foundation
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY
# KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""SCons.Tool.mslink

Tool-specific initialization for the Microsoft linker.

There normally shouldn't be any need to import this module directly.
It will usually be imported through the generic SCons.Tool.Tool()
selection method.

"""

import os
import os.path

import SCons.Action
import SCons.Defaults
import SCons.Errors
import SCons.Platform.win32
import SCons.Tool
import SCons.Tool.msvc
import SCons.Tool.msvs
import SCons.Util

from SCons.Tool.MSCommon import (
    msvc_setup_env_once,
    msvc_setup_env_tool,
)
from SCons.Tool.MSCommon.common import get_pch_node

tool_name = 'mslink'

def pdbGenerator(env, target, source, for_signature):
    try:
        return ['/PDB:%s' % target[0].attributes.pdb, '/DEBUG']
    except (AttributeError, IndexError):
        return None

def _dllTargets(target, source, env, for_signature, paramtp):
    listCmd = []
    dll = env.FindIxes(target, '%sPREFIX' % paramtp, '%sSUFFIX' % paramtp)
    if dll: listCmd.append("/out:%s"%dll.get_string(for_signature))

    implib = env.FindIxes(target, 'LIBPREFIX', 'LIBSUFFIX')
    if implib: listCmd.append("/implib:%s"%implib.get_string(for_signature))

    return listCmd

def _dllSources(target, source, env, for_signature, paramtp):
    listCmd = []

    deffile = env.FindIxes(source, "WINDOWSDEFPREFIX", "WINDOWSDEFSUFFIX")
    for src in source:
        # Check explicitly for a non-None deffile so that the __cmp__
        # method of the base SCons.Util.Proxy class used for some Node
        # proxies doesn't try to use a non-existent __dict__ attribute.
        if deffile and src == deffile:
            # Treat this source as a .def file.
            listCmd.append("/def:%s" % src.get_string(for_signature))
        else:
            # Just treat it as a generic source file.
            listCmd.append(src)
    return listCmd

def windowsShlinkTargets(target, source, env, for_signature):
    return _dllTargets(target, source, env, for_signature, 'SHLIB')

def windowsShlinkSources(target, source, env, for_signature):
    return _dllSources(target, source, env, for_signature, 'SHLIB')

def _windowsLdmodTargets(target, source, env, for_signature):
    """Get targets for loadable modules."""
    return _dllTargets(target, source, env, for_signature, 'LDMODULE')

def _windowsLdmodSources(target, source, env, for_signature):
    """Get sources for loadable modules."""
    return _dllSources(target, source, env, for_signature, 'LDMODULE')

def _dllEmitter(target, source, env, paramtp):
    """Common implementation of dll emitter."""
    SCons.Tool.msvc.validate_vars(env)

    extratargets = []
    extrasources = []

    dll = env.FindIxes(target, '%sPREFIX' % paramtp, '%sSUFFIX' % paramtp)
    no_import_lib = env.get('no_import_lib', 0)

    if not dll:
        raise SCons.Errors.UserError('A shared library should have exactly one target with the suffix: %s' % env.subst('$%sSUFFIX' % paramtp))

    insert_def = env.subst("$WINDOWS_INSERT_DEF")
    if insert_def not in ['', '0', 0] and \
       not env.FindIxes(source, "WINDOWSDEFPREFIX", "WINDOWSDEFSUFFIX"):

        # append a def file to the list of sources
        extrasources.append(
            env.ReplaceIxes(dll,
                            '%sPREFIX' % paramtp, '%sSUFFIX' % paramtp,
                            "WINDOWSDEFPREFIX", "WINDOWSDEFSUFFIX"))

    version_num, suite = SCons.Tool.msvs.msvs_parse_version(env.get('MSVS_VERSION', '6.0'))
    if version_num >= 8.0 and \
            (env.get('WINDOWS_INSERT_MANIFEST', 0) or env.get('WINDOWS_EMBED_MANIFEST', 0)):
        # MSVC 8 and above automatically generate .manifest files that must be installed
        extratargets.append(
            env.ReplaceIxes(dll,
                            '%sPREFIX' % paramtp, '%sSUFFIX' % paramtp,
                            "WINDOWSSHLIBMANIFESTPREFIX", "WINDOWSSHLIBMANIFESTSUFFIX"))

    if 'PDB' in env and env['PDB']:
        pdb = env.arg2nodes('$PDB', target=target, source=source)[0]
        extratargets.append(pdb)
        target[0].attributes.pdb = pdb

    pch_node = get_pch_node(env, target, source)
    if version_num >= 11.0 and pch_node:
        # MSVC 11 and above need the PCH object file to be added to the link line,
        # otherwise you get link error LNK2011.
        pchobj = SCons.Util.splitext(str(pch_node))[0] + '.obj'
        # print "prog_emitter, version %s, appending pchobj %s"%(version_num, pchobj)
        if pchobj not in extrasources:
            extrasources.append(pchobj)

    if not no_import_lib and \
       not env.FindIxes(target, "LIBPREFIX", "LIBSUFFIX"):
        # Append an import library to the list of targets.
        extratargets.append(
            env.ReplaceIxes(dll,
                            '%sPREFIX' % paramtp, '%sSUFFIX' % paramtp,
                            "LIBPREFIX", "LIBSUFFIX"))
        # and .exp file is created if there are exports from a DLL
        extratargets.append(
            env.ReplaceIxes(dll,
                            '%sPREFIX' % paramtp, '%sSUFFIX' % paramtp,
                            "WINDOWSEXPPREFIX", "WINDOWSEXPSUFFIX"))

    return (target+extratargets, source+extrasources)

def windowsLibEmitter(target, source, env):
    return _dllEmitter(target, source, env, 'SHLIB')

def ldmodEmitter(target, source, env):
    """Emitter for loadable modules.

    Loadable modules are identical to shared libraries on Windows, but building
    them is subject to different parameters (LDMODULE*).
    """
    return _dllEmitter(target, source, env, 'LDMODULE')

def prog_emitter(target, source, env):
    SCons.Tool.msvc.validate_vars(env)

    extratargets = []
    extrasources = []

    exe = env.FindIxes(target, "PROGPREFIX", "PROGSUFFIX")
    if not exe:
        raise SCons.Errors.UserError("An executable should have exactly one target with the suffix: %s" % env.subst("$PROGSUFFIX"))

    version_num, suite = SCons.Tool.msvs.msvs_parse_version(env.get('MSVS_VERSION', '6.0'))
    if version_num >= 8.0 and \
            (env.get('WINDOWS_INSERT_MANIFEST', 0) or env.get('WINDOWS_EMBED_MANIFEST', 0)):
        # MSVC 8 and above automatically generate .manifest files that have to be installed
        extratargets.append(
            env.ReplaceIxes(exe,
                            "PROGPREFIX", "PROGSUFFIX",
                            "WINDOWSPROGMANIFESTPREFIX", "WINDOWSPROGMANIFESTSUFFIX"))

    if 'PDB' in env and env['PDB']:
        pdb = env.arg2nodes('$PDB', target=target, source=source)[0]
        extratargets.append(pdb)
        target[0].attributes.pdb = pdb

    pch_node = get_pch_node(env, target, source)
    if version_num >= 11.0 and pch_node:
        # MSVC 11 and above need the PCH object file to be added to the link line,
        # otherwise you get link error LNK2011.
        pchobj = SCons.Util.splitext(str(pch_node))[0] + '.obj'
        # print "prog_emitter, version %s, appending pchobj %s"%(version_num, pchobj)
        if pchobj not in extrasources:
            extrasources.append(pchobj)

    return (target+extratargets,source+extrasources)

def RegServerFunc(target, source, env):
    if 'register' in env and env['register']:
        ret = regServerAction([target[0]], [source[0]], env)
        if ret:
            raise SCons.Errors.UserError("Unable to register %s" % target[0])
        else:
            print("Registered %s sucessfully" % target[0])
        return ret
    return 0

# These are the actual actions run to embed the manifest.
# They are only called from the Check versions below.
embedManifestExeAction = SCons.Action.Action('$MTEXECOM')
embedManifestDllAction = SCons.Action.Action('$MTSHLIBCOM')

def embedManifestDllCheck(target, source, env):
    """Function run by embedManifestDllCheckAction to check for existence of manifest
    and other conditions, and embed the manifest by calling embedManifestDllAction if so."""
    if env.get('WINDOWS_EMBED_MANIFEST', 0):
        manifestSrc = target[0].get_abspath() + '.manifest'
        if os.path.exists(manifestSrc):
            ret = embedManifestDllAction([target[0]], None, env)
            if ret:
                raise SCons.Errors.UserError("Unable to embed manifest into %s" % (target[0]))
            return ret
        else:
            print('(embed: no %s.manifest found; not embedding.)'%str(target[0]))
    return 0

def embedManifestExeCheck(target, source, env):
    """Function run by embedManifestExeCheckAction to check for existence of manifest
    and other conditions, and embed the manifest by calling embedManifestExeAction if so."""
    if env.get('WINDOWS_EMBED_MANIFEST', 0):
        manifestSrc = target[0].get_abspath() + '.manifest'
        if os.path.exists(manifestSrc):
            ret = embedManifestExeAction([target[0]], None, env)
            if ret:
                raise SCons.Errors.UserError("Unable to embed manifest into %s" % (target[0]))
            return ret
        else:
            print('(embed: no %s.manifest found; not embedding.)'%str(target[0]))
    return 0

embedManifestDllCheckAction = SCons.Action.Action(embedManifestDllCheck, None)
embedManifestExeCheckAction = SCons.Action.Action(embedManifestExeCheck, None)

regServerAction = SCons.Action.Action("$REGSVRCOM", "$REGSVRCOMSTR")
regServerCheck = SCons.Action.Action(RegServerFunc, None)
shlibLinkAction = SCons.Action.Action('${TEMPFILE("$SHLINK $SHLINKFLAGS $_SHLINK_TARGETS $_LIBDIRFLAGS $_LIBFLAGS $_PDB $_SHLINK_SOURCES", "$SHLINKCOMSTR")}', '$SHLINKCOMSTR')
compositeShLinkAction = shlibLinkAction + regServerCheck + embedManifestDllCheckAction
ldmodLinkAction = SCons.Action.Action('${TEMPFILE("$LDMODULE $LDMODULEFLAGS $_LDMODULE_TARGETS $_LIBDIRFLAGS $_LIBFLAGS $_PDB $_LDMODULE_SOURCES", "$LDMODULECOMSTR")}', '$LDMODULECOMSTR')
compositeLdmodAction = ldmodLinkAction + regServerCheck + embedManifestDllCheckAction
exeLinkAction = SCons.Action.Action('${TEMPFILE("$LINK $LINKFLAGS /OUT:$TARGET.windows $_LIBDIRFLAGS $_LIBFLAGS $_PDB $SOURCES.windows", "$LINKCOMSTR")}', '$LINKCOMSTR')
compositeLinkAction = exeLinkAction + embedManifestExeCheckAction

def generate(env):
    """Add Builders and construction variables for ar to an Environment."""
    SCons.Tool.createSharedLibBuilder(env, shlib_suffix='$SHLIBSUFFIX')
    SCons.Tool.createProgBuilder(env)

    env['SHLINK'] = '$LINK'
    env['SHLINKFLAGS'] = SCons.Util.CLVar('$LINKFLAGS /dll')
    env['_SHLINK_TARGETS'] = windowsShlinkTargets
    env['_SHLINK_SOURCES'] = windowsShlinkSources
    env['SHLINKCOM'] = compositeShLinkAction
    env.Append(SHLIBEMITTER=[windowsLibEmitter])
    env.Append(LDMODULEEMITTER=[windowsLibEmitter])
    env['LINK'] = 'link'
    env['LINKFLAGS'] = SCons.Util.CLVar('/nologo')
    env['_PDB'] = pdbGenerator
    env['LINKCOM'] = compositeLinkAction
    env.Append(PROGEMITTER=[prog_emitter])
    env['LIBDIRPREFIX'] = '/LIBPATH:'
    env['LIBDIRSUFFIX'] = ''
    env['LIBLINKPREFIX'] = ''
    env['LIBLINKSUFFIX'] = '$LIBSUFFIX'

    env['WINDOWSDEFPREFIX'] = ''
    env['WINDOWSDEFSUFFIX'] = '.def'
    env['WINDOWSEXPPREFIX'] = ''
    env['WINDOWSEXPSUFFIX'] = '.exp'
    env['WINDOWS_INSERT_DEF'] = 0

    env['WINDOWSSHLIBMANIFESTPREFIX'] = ''
    env['WINDOWSSHLIBMANIFESTSUFFIX'] = '${SHLIBSUFFIX}.manifest'
    env['WINDOWSPROGMANIFESTPREFIX'] = ''
    env['WINDOWSPROGMANIFESTSUFFIX'] = '${PROGSUFFIX}.manifest'

    env['REGSVRACTION'] = regServerCheck
    env['REGSVR'] = os.path.join(
        SCons.Platform.win32.get_system_root(), 'System32', 'regsvr32'
    )
    env['REGSVRFLAGS'] = '/s '
    env['REGSVRCOM'] = '$REGSVR $REGSVRFLAGS ${TARGET.windows}'

    env['WINDOWS_EMBED_MANIFEST'] = 0
    env['MT'] = 'mt'
    # env['MTFLAGS'] = ['-hashupdate']
    env['MTFLAGS'] = SCons.Util.CLVar('/nologo')
    # Note: use - here to prevent build failure if no manifest produced.
    # This seems much simpler than a fancy system using a function action to see
    # if the manifest actually exists before trying to run mt with it.
    env['MTEXECOM'] = '-$MT $MTFLAGS -manifest ${TARGET}.manifest $_MANIFEST_SOURCES -outputresource:$TARGET;1'
    env['MTSHLIBCOM'] = '-$MT $MTFLAGS -manifest ${TARGET}.manifest $_MANIFEST_SOURCES -outputresource:$TARGET;2'
    # TODO Future work garyo 27-Feb-11
    env['_MANIFEST_SOURCES'] = None  # _windowsManifestSources

    # Set-up ms tools paths
    msvc_setup_env_once(env, tool=tool_name)

    # Loadable modules are on Windows the same as shared libraries, but they
    # are subject to different build parameters (LDMODULE* variables).
    # Therefore LDMODULE* variables correspond as much as possible to
    # SHLINK*/SHLIB* ones.
    SCons.Tool.createLoadableModuleBuilder(env, loadable_module_suffix='$LDMODULESUFFIX')
    env['LDMODULE'] = '$SHLINK'
    env['LDMODULEPREFIX'] = '$SHLIBPREFIX'
    env['LDMODULESUFFIX'] = '$SHLIBSUFFIX'
    env['LDMODULEFLAGS'] = '$SHLINKFLAGS'
    env['_LDMODULE_TARGETS'] = _windowsLdmodTargets
    env['_LDMODULE_SOURCES'] = _windowsLdmodSources
    env['LDMODULEEMITTER'] = [ldmodEmitter]
    env['LDMODULECOM'] = compositeLdmodAction

    # Issue #3350
    # Change tempfile argument joining character from a space to a newline
    # mslink will fail if any single line is too long, but is fine with many lines
    # in a tempfile
    env['TEMPFILEARGJOIN'] = os.linesep

def exists(env):
    return msvc_setup_env_tool(env, tool=tool_name)

# Local Variables:
# tab-width:4
# indent-tabs-mode:nil
# End:
# vim: set expandtab tabstop=4 shiftwidth=4:



#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Code to generate and interact with compiled asyncgen objects."""

from .CodeHelpers import generateStatementSequenceCode
from .CodeObjectCodes import getCodeObjectAccessCode
from .Emission import SourceCodeCollector
from .FunctionCodes import (
    finalizeFunctionLocalVariables,
    getClosureCopyCode,
    getFunctionCreationArgs,
    getFunctionQualnameObj,
    setupFunctionLocalVariables,
)
from .Indentation import indented
from .ModuleCodes import getModuleAccessCode
from .templates.CodeTemplatesAsyncgens import (
    template_asyncgen_exception_exit,
    template_asyncgen_no_exception_exit,
    template_asyncgen_object_body,
    template_asyncgen_object_maker_template,
    template_asyncgen_return_exit,
    template_make_asyncgen,
)
from .YieldCodes import getYieldReturnDispatchCode


def _getAsyncgenMakerIdentifier(function_identifier):
    return "MAKE_ASYNCGEN_" + function_identifier


def getAsyncgenObjectDeclCode(function_identifier, closure_variables):
    asyncgen_creation_args = getFunctionCreationArgs(
        defaults_name=None,
        kw_defaults_name=None,
        annotations_name=None,
        closure_variables=closure_variables,
    )

    return template_asyncgen_object_maker_template % {
        "asyncgen_maker_identifier": _getAsyncgenMakerIdentifier(function_identifier),
        "asyncgen_creation_args": ", ".join(asyncgen_creation_args),
    }


def getAsyncgenObjectCode(
    context,
    function_identifier,
    closure_variables,
    user_variables,
    outline_variables,
    temp_variables,
    needs_exception_exit,
    needs_generator_return,
):
    # A bit of details going on here, pylint: disable=too-many-locals

    setupFunctionLocalVariables(
        context=context,
        parameters=None,
        closure_variables=closure_variables,
        user_variables=user_variables + outline_variables,
        temp_variables=temp_variables,
    )

    function_codes = SourceCodeCollector()

    asyncgen_object_body = context.getOwner()

    generateStatementSequenceCode(
        statement_sequence=asyncgen_object_body.subnode_body,
        allow_none=True,
        emit=function_codes,
        context=context,
    )

    function_cleanup = finalizeFunctionLocalVariables(context)

    if needs_exception_exit:
        (
            exception_state_name,
            _exception_lineno,
        ) = context.variable_storage.getExceptionVariableDescriptions()

        generator_exit = template_asyncgen_exception_exit % {
            "function_cleanup": indented(function_cleanup),
            "exception_state_name": exception_state_name,
        }
    else:
        generator_exit = template_asyncgen_no_exception_exit % {
            "function_cleanup": indented(function_cleanup)
        }

    if needs_generator_return:
        generator_exit += template_asyncgen_return_exit % {}

    function_locals = context.variable_storage.makeCFunctionLevelDeclarations()

    local_type_decl = context.variable_storage.makeCStructLevelDeclarations()
    function_locals += context.variable_storage.makeCStructInits()

    if local_type_decl:
        heap_declaration = """\
struct %(function_identifier)s_locals *asyncgen_heap = \
(struct %(function_identifier)s_locals *)asyncgen->m_heap_storage;""" % {
            "function_identifier": function_identifier
        }
    else:
        heap_declaration = ""

    asyncgen_creation_args = getFunctionCreationArgs(
        defaults_name=None,
        kw_defaults_name=None,
        annotations_name=None,
        closure_variables=closure_variables,
    )

    return template_asyncgen_object_body % {
        "function_identifier": function_identifier,
        "function_body": indented(function_codes.codes),
        "heap_declaration": indented(heap_declaration),
        "has_heap_declaration": 1 if heap_declaration != "" else 0,
        "function_local_types": indented(local_type_decl),
        "function_var_inits": indented(function_locals),
        "function_dispatch": indented(getYieldReturnDispatchCode(context)),
        "asyncgen_maker_identifier": _getAsyncgenMakerIdentifier(function_identifier),
        "asyncgen_creation_args": ", ".join(asyncgen_creation_args),
        "asyncgen_exit": generator_exit,
        "asyncgen_module": getModuleAccessCode(context),
        "asyncgen_name_obj": context.getConstantCode(
            constant=asyncgen_object_body.getFunctionName()
        ),
        "asyncgen_qualname_obj": getFunctionQualnameObj(asyncgen_object_body, context),
        "code_identifier": getCodeObjectAccessCode(
            code_object=asyncgen_object_body.getCodeObject(), context=context
        ),
        "closure_name": "closure" if closure_variables else "NULL",
        "closure_count": len(closure_variables),
    }


def generateMakeAsyncgenObjectCode(to_name, expression, emit, context):
    asyncgen_object_body = expression.subnode_asyncgen_ref.getFunctionBody()

    closure_variables = expression.getClosureVariableVersions()

    closure_name, closure_copy = getClosureCopyCode(
        closure_variables=closure_variables, context=context
    )

    args = ["tstate"]
    if closure_name:
        args.append(closure_name)

    emit(
        template_make_asyncgen
        % {
            "to_name": to_name,
            "asyncgen_maker_identifier": _getAsyncgenMakerIdentifier(
                asyncgen_object_body.getCodeName()
            ),
            "args": ", ".join(str(arg) for arg in args),
            "closure_copy": indented(closure_copy),
        }
    )

    context.addCleanupTempName(to_name)


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.gnu.org/licenses/agpl.txt
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


from __future__ import print_function

import pprint

print("Importing child.")


def displayDict(d):
    d = dict(d)

    if "__loader__" in d:
        d["__loader__"] = "<__loader__ removed>"

    if "__file__" in d:
        d["__file__"] = "<__file__ removed>"

    if "__firstlineno__" in d:
        del d["__firstlineno__"]

    if "__static_attributes__" in d:
        del d["__static_attributes__"]

    # Avoid recursion that we don't offer for classes.
    if "__locals__" in d:
        del d["__locals__"]

    return pprint.pformat(d)


class A:
    pass


print("Class defined here, has these vars", displayDict(vars(A)))

#     Python tests originally created or extracted from other peoples work. The
#     parts were too small to be protected.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Trace collection (also often still referred to as constraint collection).

At the core of value propagation there is the collection of constraints that
allow to propagate knowledge forward or not.

This is about collecting these constraints and to manage them.
"""

import contextlib
from collections import defaultdict
from contextlib import contextmanager

from nuitka.__past__ import iterItems  # Python3 compatibility.
from nuitka.containers.OrderedDicts import OrderedDict
from nuitka.containers.OrderedSets import OrderedSet
from nuitka.ModuleRegistry import addUsedModule
from nuitka.nodes.NodeMakingHelpers import getComputationResult
from nuitka.nodes.shapes.StandardShapes import tshape_uninitialized
from nuitka.States import states
from nuitka.Tracing import (
    inclusion_logger,
    printError,
    printLine,
    printSeparator,
)
from nuitka.tree.SourceHandling import readSourceLine
from nuitka.utils.InstanceCounters import (
    counted_del,
    counted_init,
    isCountingInstances,
)
from nuitka.utils.Timing import TimerReport

from .ValueTraces import (
    ValueTraceAssign,
    ValueTraceAssignUnescapable,
    ValueTraceAssignUnescapablePropagated,
    ValueTraceAssignVeryTrusted,
    ValueTraceDeleted,
    ValueTraceEscaped,
    ValueTraceLoopComplete,
    ValueTraceLoopIncomplete,
    ValueTraceMerge,
    ValueTraceStartInit,
    ValueTraceStartInitStarArgs,
    ValueTraceStartInitStarDict,
    ValueTraceStartUninitialized,
    ValueTraceStartUnknown,
    ValueTraceUnknown,
)

# Keeping trace of how often branches are merged between calls
_merge_counts = defaultdict(int)


def fetchMergeCounts():
    result = dict(_merge_counts)
    _merge_counts.clear()
    return result


signalChange = None


@contextmanager
def withChangeIndicationsTo(signal_change):
    """Decide where change indications should go to."""

    global signalChange  # Singleton, pylint: disable=global-statement

    old = signalChange
    signalChange = signal_change
    yield
    signalChange = old


class CollectionUpdateMixin(object):
    """Mixin to use in every collection to add traces."""

    # Mixins are not allowed to specify slots.
    __slots__ = ()

    def getVariableTracesAll(self):
        return self.variable_traces

    def addVariableMergeMultipleTrace(self, variable, traces):
        trace_merge = ValueTraceMerge(traces)

        version = variable.allocateTargetNumber()

        self.variable_traces[variable][version] = trace_merge

        return version


class CollectionStartPointMixin(CollectionUpdateMixin):
    """Mixin to use in start points of collections.

    These are modules, functions, etc. typically entry points.
    """

    # Many things are traced, pylint: disable=too-many-instance-attributes

    # Mixins are not allowed to specify slots, pylint: disable=assigning-non-slot
    __slots__ = ()

    def __init__(self):
        # Variable assignments performed in here, last issued number, only used
        # to determine the next number that should be used for a new assignment.
        self.variable_versions = {}

        # The full trace of a variable with a version for the function or module
        # this is.
        self.variable_traces = defaultdict(dict)

        self.break_collections = None
        self.continue_collections = None
        self.return_collections = None
        self.exception_collections = None

        self.outline_functions = None

        # What loop variables were there, them going away is something we want
        # to know.
        self.loop_variables = set()

        self.delayed_work = []

    def getLoopBreakCollections(self):
        return self.break_collections

    def onLoopBreak(self, collection):
        self.break_collections.append(
            TraceCollectionSnapshot(parent=collection, name="loop break")
        )

    def getLoopContinueCollections(self):
        return self.continue_collections

    def onLoopContinue(self, collection):
        self.continue_collections.append(
            TraceCollectionSnapshot(parent=collection, name="loop continue")
        )

    def onFunctionReturn(self, collection):
        if self.return_collections is not None:
            self.return_collections.append(
                TraceCollectionSnapshot(parent=collection, name="return")
            )

    def onExceptionRaiseExit(self, raisable_exceptions, collection=None):
        """Indicate to the trace collection what exceptions may have occurred.

        Args:
            raisable_exception: Currently ignored, one or more exceptions that
            could occur, e.g. "BaseException".
            collection: To pass the collection that will be the parent
        Notes:
            Currently this is unused. Passing "collection" as an argument, so
            we know the original collection to attach the branch to, is maybe
            not the most clever way to do this

            We also might want to specialize functions for specific exceptions,
            there is little point in providing BaseException as an argument in
            so many places.

            The actual storage of the exceptions that can occur is currently
            missing entirely. We just use this to detect "any exception" by
            not being empty.
        """

        # TODO: We might want to track per exception, pylint: disable=unused-argument

        if self.exception_collections is not None:
            if collection is None:
                collection = self

            if not self.exception_collections or (
                self.exception_collections[-1].variable_actives
                is not collection.variable_actives
            ):
                self.exception_collections.append(
                    TraceCollectionSnapshot(parent=collection, name="exception")
                )

    def getFunctionReturnCollections(self):
        return self.return_collections

    def getExceptionRaiseCollections(self):
        return self.exception_collections

    @contextlib.contextmanager
    def makeAbortStackContext(
        self, catch_breaks, catch_continues, catch_returns, catch_exceptions
    ):
        if catch_breaks:
            old_break_collections = self.break_collections
            self.break_collections = []
        if catch_continues:
            old_continue_collections = self.continue_collections
            self.continue_collections = []
        if catch_returns:
            old_return_collections = self.return_collections
            self.return_collections = []
        if catch_exceptions:
            old_exception_collections = self.exception_collections
            self.exception_collections = []

        yield

        if catch_breaks:
            self.break_collections = old_break_collections
        if catch_continues:
            self.continue_collections = old_continue_collections
        if catch_returns:
            self.return_collections = old_return_collections
        if catch_exceptions:
            self.exception_collections = old_exception_collections

    def addOutlineFunction(self, outline):
        if self.outline_functions is None:
            self.outline_functions = [outline]
        else:
            self.outline_functions.append(outline)

    def getOutlineFunctions(self):
        return self.outline_functions

    def onLocalsDictEscaped(self, locals_scope):
        locals_scope.preventLocalsDictPropagation()

        for variable in locals_scope.variables.values():
            self.markActiveVariableAsEscaped(variable)

        # TODO: Limit to the scope.
        # TODO: Does the above code not do that already?
        for variable in self.variable_actives:
            if variable.isTempVariable() or variable.isModuleVariable():
                continue

            self.markActiveVariableAsEscaped(variable)

    def onUsedFunction(self, function_body):
        owning_module = function_body.getParentModule()

        # Make sure the owning module is added to the used set. This is most
        # important for helper functions, or modules, which otherwise have
        # become unused.
        addUsedModule(
            module=owning_module,
            using_module=None,
            usage_tag="function",
            reason="Function %s" % self.name,
            source_ref=owning_module.source_ref,
        )

        needs_visit = owning_module.addUsedFunction(function_body)

        if needs_visit or function_body.isExpressionFunctionPureBody():
            function_body.computeFunctionRaw(self)

    def onDelayedWork(self, node, function, old_desc, describe_new_node):
        self.delayed_work.append((node, function, old_desc, describe_new_node))

    def performDelayedWork(self):
        for node, delayed_function, old_desc, describe_new_node in self.delayed_work:
            # We might point to a node that has been removed.
            if not node.isConnected():
                if states.is_debug:
                    # Might be normal though, but we will aim at avoiding the
                    # necessity of encountering this where possible.
                    assert False, "Node %s is not connected" % node
                continue

            new_node = delayed_function(node)
            if new_node is not node:
                node.getParent().replaceChild(node, new_node)

                tags, new_node_description = describe_new_node(new_node)
                self.signalChange(
                    tags=tags,
                    message="Replaced %s with %s." % (old_desc, new_node_description),
                    source_ref=node.getSourceReference(),
                )

        # Surely do this only once.
        del self.delayed_work

    def updateVeryTrustedModuleVariable(self, variable, old_node, new_node):
        """Update a very trusted module variable assignment if necessary."""
        if self.very_trusted_module_variables.get(variable, None) is old_node:
            self.very_trusted_module_variables[variable] = new_node


class TraceCollectionBase(object):
    """This contains for logic for maintaining active traces.

    They are kept for "variable" and versions.
    """

    __slots__ = (
        "owner",
        "parent",
        "name",
        "variable_actives",
        "variable_actives_needs_copy",
        "has_unescaped_variables",
        "variable_escapable",
    )

    if isCountingInstances():
        __del__ = counted_del()

    @counted_init
    def __init__(self, owner, name, parent):
        self.owner = owner
        self.parent = parent
        self.name = name

        # Currently active values in the tracing.
        self.variable_actives = {}
        self.variable_actives_needs_copy = True

        # Even though it's empty, we set it, because init of variables won't do it.
        self.has_unescaped_variables = True

        self.variable_escapable = set()

    def __repr__(self):
        return "<%s for %s at 0x%x>" % (self.__class__.__name__, self.name, id(self))

    def getOwner(self):
        return self.owner

    def dumpActiveTraces(self, indent=""):
        printSeparator()
        printLine("Active are:")
        for variable, version in sorted(
            self.variable_actives.items(), key=lambda var: var[0].variable_name
        ):
            printLine("%s %s:" % (variable, version))
            self.getVariableCurrentTrace(variable).dump(indent)

        printSeparator()

    def getVariableCurrentTrace(self, variable):
        """Get the current value trace associated to this variable

        It is also created on the fly if necessary. We create them
        lazy so to keep the tracing branches minimal where possible.
        """

        return self.variable_traces[variable][self.variable_actives[variable]]

    def hasVariableCurrentTrace(self, variable):
        return variable in self.variable_actives

    def markCurrentVariableTrace(self, variable, version):
        if self.variable_actives_needs_copy:
            self.variable_actives = self.variable_actives.copy()
            self.variable_actives_needs_copy = False

        self.variable_actives[variable] = version
        self.has_unescaped_variables = True

    def removeCurrentVariableTrace(self, variable):
        if self.variable_actives_needs_copy:
            self.variable_actives = self.variable_actives.copy()
            self.variable_actives_needs_copy = False

        del self.variable_actives[variable]

    def initVariableLate(self, variable):
        if self.variable_actives_needs_copy:
            self.variable_actives = self.variable_actives.copy()
            self.variable_actives_needs_copy = False

        variable.initVariableLate(self)

    def markActiveVariableAsEscaped(self, variable):
        version = self.variable_actives[variable]
        variable_traces = self.variable_traces[variable]
        current = variable_traces[version]

        if current.isTraceThatNeedsEscape():
            # Escape traces are div 3 rem 1.
            version = version // 3 * 3 + 1

            if version not in variable_traces:
                variable_traces[version] = ValueTraceEscaped(self.owner, current)

            self.markCurrentVariableTrace(variable, version)

    def markClosureVariableAsUnknown(self, variable):
        version = self.variable_actives[variable]
        variable_traces = self.variable_traces[variable]
        current = variable_traces[version]

        if not current.isUnknownTrace():
            # Unknown traces are div 3 rem 2.
            version = version // 3 * 3 + 2

            if version not in variable_traces:
                variable_traces[version] = ValueTraceUnknown(self.owner, current)

            self.markCurrentVariableTrace(variable, version)

    def markActiveVariableAsUnknown(self, variable):
        version = self.variable_actives[variable]

        if version % 3 != 2:
            variable_traces = self.variable_traces[variable]
            current = variable_traces[version]

            if not current.isUnknownOrVeryTrustedTrace():
                # Unknown traces are div 3 rem 2.
                version = version // 3 * 3 + 2

                if version not in variable_traces:
                    variable_traces[version] = ValueTraceUnknown(self.owner, current)

                self.markCurrentVariableTrace(variable, version)

    def markActiveVariableAsLoopMerge(
        self, loop_node, current, variable, shapes, incomplete
    ):
        if incomplete:
            result = ValueTraceLoopIncomplete(loop_node, current, shapes)
        else:
            # TODO: Empty is a missing optimization somewhere, but it also happens that
            # a variable is getting released in a loop.
            # assert shapes, (variable, current)

            if not shapes:
                shapes.add(tshape_uninitialized)

            result = ValueTraceLoopComplete(loop_node, current, shapes)

        version = variable.allocateTargetNumber()
        self.variable_traces[variable][version] = result

        self.markCurrentVariableTrace(variable, version)

        return result

    @staticmethod
    def signalChange(tags, source_ref, message):
        # This is monkey patched from another module. pylint: disable=I0021,not-callable
        signalChange(tags, source_ref, message)

    @staticmethod
    def mustAlias(a, b):
        if a.isExpressionVariableRef() and b.isExpressionVariableRef():
            return a.getVariable() is b.getVariable()

        return False

    @staticmethod
    def mustNotAlias(a, b):
        # TODO: not yet really implemented
        if a.isExpressionConstantRef() and b.isExpressionConstantRef():
            if a.isMutable() or b.isMutable():
                return True

        return False

    def onControlFlowEscape(self, node):
        # TODO: One day, we should trace which nodes exactly cause a variable
        # to be considered escaped, pylint: disable=unused-argument

        if self.has_unescaped_variables:
            #            print("Control flow escape in", self.name, self.variable_escapable)
            for variable in self.variable_escapable:
                if variable in self.variable_actives:
                    variable.onControlFlowEscape(self)

            self.has_unescaped_variables = False

    def removeKnowledge(self, node):
        if node.isExpressionVariableRef():
            node.variable.removeKnowledge(self)

    def onValueEscapeStr(self, node):
        # TODO: We can ignore these for now.
        pass

    def removeAllKnowledge(self):
        for variable in self.variable_actives:
            variable.removeAllKnowledge(self)

    def onVariableSet(self, variable, version, assign_node):
        variable_trace = ValueTraceAssign(
            self.owner,
            assign_node,
            self.getVariableCurrentTrace(variable),
        )

        self.variable_traces[variable][version] = variable_trace

        # Make references point to it.
        self.markCurrentVariableTrace(variable, version)

        return variable_trace

    def onVariableSetAliasing(self, variable, version, assign_node, source):
        other_variable_trace = source.variable_trace

        if other_variable_trace.__class__ is ValueTraceAssignUnescapable:
            return self.onVariableSetToUnescapableValue(
                variable=variable, version=version, assign_node=assign_node
            )
        elif other_variable_trace.__class__ is ValueTraceAssignVeryTrusted:
            return self.onVariableSetToVeryTrustedValue(
                variable=variable, version=version, assign_node=assign_node
            )
        else:
            result = self.onVariableSet(
                variable=variable, version=version, assign_node=assign_node
            )

            self.removeKnowledge(source)

            return result

    def onVariableSetToUnescapableValue(self, variable, version, assign_node):
        variable_trace = ValueTraceAssignUnescapable(
            self.owner,
            assign_node,
            self.getVariableCurrentTrace(variable),
        )

        self.variable_traces[variable][version] = variable_trace

        # Make references point to it.
        self.markCurrentVariableTrace(variable, version)

        return variable_trace

    def onVariableSetToVeryTrustedValue(self, variable, version, assign_node):
        variable_trace = ValueTraceAssignVeryTrusted(
            self.owner,
            assign_node,
            self.getVariableCurrentTrace(variable),
        )

        self.variable_traces[variable][version] = variable_trace

        # Make references point to it.
        self.markCurrentVariableTrace(variable, version)

        return variable_trace

    def onVariableSetToUnescapablePropagatedValue(
        self, variable, version, assign_node, replacement
    ):
        variable_trace = ValueTraceAssignUnescapablePropagated(
            self.owner,
            assign_node,
            self.getVariableCurrentTrace(variable),
            replacement,
        )

        self.variable_traces[variable][version] = variable_trace

        # Make references point to it.
        self.markCurrentVariableTrace(variable, version)

        return variable_trace

    def onVariableDel(self, variable, version, del_node):
        # Add a new trace, allocating a new version for the variable, and
        # remember the delete of the current
        old_trace = self.getVariableCurrentTrace(variable)

        # TODO: Annotate value content as escaped.

        variable_trace = ValueTraceDeleted(
            self.owner,
            old_trace,
            del_node,
        )

        # Assign to not initialized again.
        self.variable_traces[variable][version] = variable_trace

        # Make references point to it.
        self.markCurrentVariableTrace(variable, version)

        return variable_trace

    def onLocalsUsage(self, locals_scope):
        self.onLocalsDictEscaped(locals_scope)

        result = []

        scope_locals_variables = locals_scope.getLocalsRelevantVariables()

        for variable in self.variable_actives:
            if variable.isLocalVariable() and variable in scope_locals_variables:
                variable_trace = self.getVariableCurrentTrace(variable)

                variable_trace.addNameUsage()
                result.append((variable, variable_trace))

        return result

    def onVariableContentEscapes(self, variable):
        self.markActiveVariableAsEscaped(variable)

    def onExpression(self, expression, allow_none=False):
        if expression is None and allow_none:
            return None

        parent = expression.parent

        # Now compute this expression, allowing it to replace itself with
        # something else as part of a local peep hole optimization.
        new_node, change_tags, change_desc = expression.computeExpressionRaw(self)

        if change_tags is not None:
            # This is mostly for tracing and indication that a change occurred
            # and it may be interesting to look again.
            self.signalChange(change_tags, expression.getSourceReference(), change_desc)

        if new_node is not expression:
            parent.replaceChild(expression, new_node)

        return new_node

    def onStatement(self, statement):
        try:
            new_statement, change_tags, change_desc = statement.computeStatement(self)

            # print new_statement, change_tags, change_desc
            if new_statement is not statement:
                self.signalChange(
                    change_tags, statement.getSourceReference(), change_desc
                )

            return new_statement
        except Exception:
            printError(
                "Problem with statement at %s:\n-> %s"
                % (
                    statement.source_ref.getAsString(),
                    readSourceLine(statement.source_ref),
                )
            )
            raise

    def computedStatementResult(self, statement, change_tags, change_desc):
        """Make sure the replacement statement is computed.

        Use this when a replacement expression needs to be seen by the trace
        collection and be computed, without causing any duplication, but where
        otherwise there would be loss of annotated effects.

        This may e.g. be true for nodes that need an initial run to know their
        exception result and type shape.
        """
        # Need to compute the replacement still.
        new_statement = statement.computeStatement(self)

        if new_statement[0] is not statement:
            # Signal intermediate result as well.
            self.signalChange(change_tags, statement.getSourceReference(), change_desc)

            return new_statement
        else:
            return statement, change_tags, change_desc

    def computedExpressionResult(self, expression, change_tags, change_desc):
        """Make sure the replacement expression is computed.

        Use this when a replacement expression needs to be seen by the trace
        collection and be computed, without causing any duplication, but where
        otherwise there would be loss of annotated effects.

        This may e.g. be true for nodes that need an initial run to know their
        exception result and type shape.
        """

        # Need to compute the replacement still.
        new_expression = expression.computeExpression(self)

        if new_expression[0] is not expression:
            # Signal intermediate result as well.
            self.signalChange(change_tags, expression.getSourceReference(), change_desc)

            return new_expression
        else:
            return expression, change_tags, change_desc

    def computedExpressionResultRaw(self, expression, change_tags, change_desc):
        """Make sure the replacement expression is computed.

        Use this when a replacement expression needs to be seen by the trace
        collection and be computed, without causing any duplication, but where
        otherwise there would be loss of annotated effects.

        This may e.g. be true for nodes that need an initial run to know their
        exception result and type shape.

        This is for raw, i.e. subnodes are not yet computed automatically.
        """

        # Need to compute the replacement still.
        new_expression = expression.computeExpressionRaw(self)

        if new_expression[0] is not expression:
            # Signal intermediate result as well.
            self.signalChange(change_tags, expression.getSourceReference(), change_desc)

            return new_expression
        else:
            return expression, change_tags, change_desc

    def mergeBranches(self, collection_yes, collection_no):
        """Merge two alternative branches into this trace.

        This is mostly for merging conditional branches, or other ways
        of having alternative control flow. This deals with up to two
        alternative branches to both change this collection.
        """

        # Many branches due to inlining the actual merge and preparing it
        # pylint: disable=too-many-branches

        if collection_yes is None:
            if collection_no is not None:
                # Handle one branch case, we need to merge versions backwards as
                # they may make themselves obsolete.
                collection1 = self
                collection2 = collection_no

                # Nothing to do here, self is unchanged.
                if collection1.variable_actives is collection2.variable_actives:
                    return
            else:
                # Refuse to do stupid work
                return
        elif collection_no is None:
            # Handle one branch case, we need to merge versions backwards as
            # they may make themselves obsolete.
            collection1 = self
            collection2 = collection_yes

            # Nothing to do here, self is unchanged.
            if collection1.variable_actives is collection2.variable_actives:
                return

        else:
            # Handle two branch case, they may or may not do the same things.
            collection1 = collection_yes
            collection2 = collection_no

            # Both branches being unchanged, means no merge is needed
            if collection1.variable_actives == collection2.variable_actives:
                self.replaceBranch(collection1)

                return

        _merge_counts[2] += 1

        if states.is_debug:
            # They must have the same content only or else some bug occurred.
            if len(collection1.variable_actives) != len(collection2.variable_actives):
                for variable, version in iterItems(collection1.variable_actives):
                    if variable not in collection2.variable_actives:
                        print("Only in collection1", variable, version)

                for variable, version in iterItems(collection2.variable_actives):
                    if variable not in collection1.variable_actives:
                        print("Only in collection2", variable, version)

                assert False

        has_unescaped_variables = (
            collection1.has_unescaped_variables or collection2.has_unescaped_variables
        )
        new_actives = {}
        for variable, version in iterItems(collection1.variable_actives):
            other_version = collection2.variable_actives[variable]

            if version != other_version:
                variable_traces = self.variable_traces[variable]

                trace1 = variable_traces[version]
                trace2 = variable_traces[other_version]

                if version % 3 == 1 and trace1.previous is trace2:
                    pass
                elif other_version % 3 == 1 and trace2.previous is trace1:
                    version = other_version
                else:
                    version = self.addVariableMergeMultipleTrace(
                        variable,
                        (
                            trace1,
                            trace2,
                        ),
                    )

                    has_unescaped_variables = True

            new_actives[variable] = version

        self.variable_actives = new_actives
        self.variable_actives_needs_copy = False

        # TODO: This could be avoided, if we detect no actual changes being present, but it might
        # be more costly.
        self.has_unescaped_variables = has_unescaped_variables

    def mergeMultipleBranches(self, collections):
        # Optimize for length 1, which is trivial merge and needs not a
        # lot of work, and length 2 has dedicated code as it's so frequent.

        # collection_levels = set()
        # new_collections = []

        # for collection in collections:
        #     if collection.variable_actives_level not in collection_levels:
        #         collection_levels.add(collection.variable_actives_level)
        #         new_collections.append(collection)

        # collections = new_collections

        # if len(collections) != len(new_collections):
        #     print("Reduced multi %d to %d for %s" % (len(collections), len(new_collections), self))
        #     assert False

        merge_size = len(collections)

        if merge_size == 1:
            self.replaceBranch(collections[0])
            return
        elif merge_size == 2:
            return self.mergeBranches(*collections)

        _merge_counts[len(collections)] += 1

        with TimerReport(
            message="Running merge for %s took %%.2f seconds" % collections,
            decider=False,
            include_sleep_time=False,
            use_perf_counters=False,
        ):
            new_actives = {}

            has_unescaped_variables = any(
                collection.has_unescaped_variables for collection in collections
            )

            for variable in collections[0].variable_actives:
                versions = set(
                    collection.variable_actives[variable] for collection in collections
                )

                if len(versions) == 1:
                    (version,) = versions
                else:
                    traces = []
                    escaped = []
                    winner_version = None

                    variable_traces = self.variable_traces[variable]

                    for version in sorted(versions):
                        trace = variable_traces[version]

                        if version % 3 == 1:
                            winner_version = version
                            escaped_trace = trace.previous

                            if escaped_trace in traces:
                                traces.remove(trace.previous)

                            escaped.append(escaped)
                            traces.append(trace)
                        else:
                            if trace not in escaped:
                                traces.append(trace)

                    if len(traces) == 1:
                        version = winner_version
                        assert winner_version is not None
                    else:
                        version = self.addVariableMergeMultipleTrace(
                            variable,
                            traces,
                        )

                        has_unescaped_variables = True

                new_actives[variable] = version

            self.variable_actives = new_actives
            self.variable_actives_needs_copy = False

            # TODO: This could be avoided, if we detect no actual changes being present, but it might
            # be more costly.
            self.has_unescaped_variables = has_unescaped_variables

    def replaceBranch(self, collection_replace):
        self.variable_actives = collection_replace.variable_actives
        self.has_unescaped_variables = collection_replace.has_unescaped_variables

        # Make the old one unusable.
        collection_replace.variable_actives = None

        _merge_counts[1] += 1

    def onLoopBreak(self, collection):
        return self.parent.onLoopBreak(collection)

    def onLoopContinue(self, collection):
        return self.parent.onLoopContinue(collection)

    def onFunctionReturn(self, collection):
        return self.parent.onFunctionReturn(collection)

    def onExceptionRaiseExit(self, raisable_exceptions, collection=None):
        if collection is None:
            collection = self

        return self.parent.onExceptionRaiseExit(raisable_exceptions, collection)

    def getLoopBreakCollections(self):
        return self.parent.getLoopBreakCollections()

    def getLoopContinueCollections(self):
        return self.parent.getLoopContinueCollections()

    def getFunctionReturnCollections(self):
        return self.parent.getFunctionReturnCollections()

    def getExceptionRaiseCollections(self):
        return self.parent.getExceptionRaiseCollections()

    def makeAbortStackContext(
        self, catch_breaks, catch_continues, catch_returns, catch_exceptions
    ):
        return self.parent.makeAbortStackContext(
            catch_breaks=catch_breaks,
            catch_continues=catch_continues,
            catch_returns=catch_returns,
            catch_exceptions=catch_exceptions,
        )

    def onLocalsDictEscaped(self, locals_scope):
        self.parent.onLocalsDictEscaped(locals_scope)

    def getCompileTimeComputationResult(
        self, node, computation, description, user_provided=False
    ):
        new_node, change_tags, message = getComputationResult(
            node=node,
            computation=computation,
            description=description,
            user_provided=user_provided,
        )

        if change_tags == "new_raise":
            self.onExceptionRaiseExit(BaseException)

        return new_node, change_tags, message

    def addOutlineFunction(self, outline):
        self.parent.addOutlineFunction(outline)

    def getVeryTrustedModuleVariables(self):
        return self.parent.getVeryTrustedModuleVariables()

    def updateVeryTrustedModuleVariable(self, variable, old_node, new_node):
        return self.parent.updateVeryTrustedModuleVariable(variable, old_node, new_node)

    def onUsedFunction(self, function_body):
        return self.parent.onUsedFunction(function_body)

    def onModuleUsageAttempts(self, module_usage_attempts):
        self.parent.onModuleUsageAttempts(module_usage_attempts)

    def onModuleUsageAttempt(self, module_usage_attempt):
        self.parent.onModuleUsageAttempt(module_usage_attempt)

    def onDistributionUsed(self, distribution_name, node, success):
        self.parent.onDistributionUsed(
            distribution_name=distribution_name, node=node, success=success
        )

    def initVariableUnknown(self, variable, old_collection):
        # TODO: Making these really reusable by removing their state for each of
        # these cases. pylint: disable=unused-argument
        # if old_collection is not None:
        #     trace = old_collection.variable_traces[variable][-1]
        # else:
        trace = ValueTraceStartUnknown(self.owner)

        # Unknown traces are div 3 rem 2.
        self.variable_traces[variable][-1] = trace
        self.variable_actives[variable] = -1

        return trace

    def initVariableModule(self, variable, old_collection):
        # TODO: Making these really reusable by removing their state for each of
        # these cases. pylint: disable=unused-argument
        # if old_collection is not None:
        #     trace = old_collection.variable_traces[variable][-1]
        # else:
        trace = ValueTraceStartUnknown(self.owner)

        # Unknown traces are div 3 rem 2.
        self.variable_traces[variable][-1] = trace
        self.variable_actives[variable] = -1

        return trace

    def initVariableInit(self, variable, old_collection):
        # TODO: Making these really reusable by removing their state for each of
        # these cases. pylint: disable=unused-argument
        # if old_collection is not None:
        #     trace = old_collection.variable_traces[variable][0]
        # else:
        trace = ValueTraceStartInit(self.owner)

        self.variable_traces[variable][0] = trace
        self.variable_actives[variable] = 0

        return trace

    def initVariableInitStarArgs(self, variable, old_collection):
        # TODO: Making these really reusable by removing their state for each of
        # these cases. pylint: disable=unused-argument
        # if old_collection is not None:
        #     trace = old_collection.variable_traces[variable][0]
        # else:
        trace = ValueTraceStartInitStarArgs(self.owner)

        self.variable_traces[variable][0] = trace
        self.variable_actives[variable] = 0

        return trace

    def initVariableInitStarDict(self, variable, old_collection):
        # TODO: Making these really reusable by removing their state for each of
        # these cases. pylint: disable=unused-argument
        # if old_collection is not None:
        #     trace = old_collection.variable_traces[variable][0]
        # else:
        trace = ValueTraceStartInitStarDict(self.owner)

        self.variable_traces[variable][0] = trace
        self.variable_actives[variable] = 0

        return trace

    def initVariableUninitialized(self, variable, old_collection):
        # TODO: Making these really reusable by removing their state for each of
        # these cases. pylint: disable=unused-argument
        # if old_collection is not None:
        #     trace = old_collection.variable_traces[variable][0]
        # else:
        trace = ValueTraceStartUninitialized(self.owner)

        self.variable_traces[variable][0] = trace
        self.variable_actives[variable] = 0

        return trace

    def initVariableUninitializedLate(self, variable):
        trace = ValueTraceStartUninitialized(self.owner)

        self.variable_traces[variable][0] = trace
        variable.setTracesForUserFirst(self.owner, self.variable_traces[variable])

        return trace

    def onDelayedWork(self, node, function, old_desc, describe_new_node):
        self.parent.onDelayedWork(node, function, old_desc, describe_new_node)


class TraceCollectionBranch(CollectionUpdateMixin, TraceCollectionBase):
    __slots__ = (
        "variable_traces",
        "loop_variables",
    )

    def __init__(self, name, parent):
        TraceCollectionBase.__init__(
            self,
            owner=parent.owner,
            name=name,
            parent=parent,
        )

        # If it gets modified, it will copy the snapshot state for us, so this
        # stays in tact.
        self.variable_actives = parent.variable_actives
        parent.variable_actives_needs_copy = True

        self.variable_escapable = parent.variable_escapable
        self.has_unescaped_variables = parent.has_unescaped_variables

        # For quick access without going to parent.
        self.variable_traces = parent.variable_traces
        self.loop_variables = parent.loop_variables

    def computeBranch(self, branch):
        result = branch.computeStatementsSequence(self)

        if result is not branch:
            branch.parent.replaceChild(branch, result)

        return result


class TraceCollectionSnapshot(CollectionUpdateMixin, TraceCollectionBase):
    __slots__ = (
        "variable_traces",
        "loop_variables",
    )

    def __init__(self, name, parent):
        TraceCollectionBase.__init__(
            self,
            owner=parent.owner,
            name=name,
            parent=parent,
        )

        # If it gets modified, it will copy the snapshot state for us, so this
        # stays in tact.
        self.variable_actives = parent.variable_actives
        parent.variable_actives_needs_copy = True

        self.variable_escapable = parent.variable_escapable
        self.has_unescaped_variables = parent.has_unescaped_variables

        # For quick access without going to parent.
        self.variable_traces = parent.variable_traces
        self.loop_variables = parent.loop_variables


class TraceCollectionFunction(CollectionStartPointMixin, TraceCollectionBase):
    __slots__ = (
        "variable_versions",
        "variable_traces",
        "loop_variables",
        "break_collections",
        "continue_collections",
        "return_collections",
        "exception_collections",
        "outline_functions",
        "very_trusted_module_variables",
        "delayed_work",
    )

    def __init__(self, parent, old_collection, function_body):
        # Many kinds of variables to setup, pylint: disable=too-many-branches

        # assert (
        #     function_body.isExpressionFunctionBody()
        #     or function_body.isExpressionGeneratorObjectBody()
        #     or function_body.isExpressionCoroutineObjectBody()
        #     or function_body.isExpressionAsyncgenObjectBody()
        # ), function_body

        CollectionStartPointMixin.__init__(self)

        TraceCollectionBase.__init__(
            self,
            owner=function_body,
            name=function_body.getCodeName(),
            parent=parent,
        )

        if parent is not None:
            self.very_trusted_module_variables = parent.getVeryTrustedModuleVariables()
        else:
            self.very_trusted_module_variables = ()

        if function_body.isExpressionFunctionBody():
            parameters = function_body.getParameters()

            for parameter_variable in parameters.getTopLevelVariables():
                self.initVariableInit(parameter_variable, old_collection)
                self.variable_escapable.add(parameter_variable)

            list_star_variable = parameters.getListStarArgVariable()
            if list_star_variable is not None:
                self.initVariableInitStarArgs(list_star_variable, old_collection)
                self.variable_escapable.add(list_star_variable)

            dict_star_variable = parameters.getDictStarArgVariable()
            if dict_star_variable is not None:
                self.initVariableInitStarDict(dict_star_variable, old_collection)
                self.variable_escapable.add(dict_star_variable)

        for closure_variable in function_body.getClosureVariables():
            if closure_variable not in self.variable_actives:
                self.initVariableUnknown(closure_variable, old_collection)

                if closure_variable.isLocalVariable():
                    self.variable_escapable.add(closure_variable)

        for local_variable in function_body.getLocalVariables():
            if local_variable not in self.variable_actives:
                self.initVariableUninitialized(local_variable, old_collection)
                self.variable_escapable.add(local_variable)

        for module_variable in function_body.getModuleVariables():
            self.initVariableModule(module_variable, old_collection)
            self.variable_escapable.add(module_variable)

        for temp_variable in function_body.getTempVariables(outline=None):
            self.initVariableUninitialized(temp_variable, old_collection)

        # TODO: Have special function type for exec functions stuff.
        locals_scope = function_body.getLocalsScope()

        if locals_scope is not None:
            if not locals_scope.isMarkedForPropagation():
                for locals_dict_variable in locals_scope.variables.values():
                    self.initVariableUninitialized(locals_dict_variable, old_collection)
            else:
                function_body.locals_scope = None

        self.has_unescaped_variables = True

    def initVariableModule(self, variable, old_collection):
        # print("initVariableModule", variable, self)
        trusted_node = self.very_trusted_module_variables.get(variable)

        if trusted_node is None:
            return TraceCollectionBase.initVariableModule(
                self, variable, old_collection
            )

        assign_trace = ValueTraceAssignVeryTrusted(
            self.owner,
            trusted_node.getParent(),
            None,
        )

        # TODO: Make very trusted assign traces also div recognized by expanding
        # to div 4.
        self.variable_traces[variable][0] = assign_trace
        self.variable_actives[variable] = 0

        return assign_trace


class TraceCollectionPureFunction(TraceCollectionFunction):
    """Pure functions don't feed their parent."""

    __slots__ = ("used_functions",)

    def __init__(self, old_collection, function_body):
        TraceCollectionFunction.__init__(
            self,
            parent=None,
            old_collection=old_collection,
            function_body=function_body,
        )

        self.used_functions = OrderedSet()

    def getUsedFunctions(self):
        return self.used_functions

    def onUsedFunction(self, function_body):
        self.used_functions.add(function_body)

        TraceCollectionFunction.onUsedFunction(self, function_body=function_body)


class TraceCollectionModule(CollectionStartPointMixin, TraceCollectionBase):
    __slots__ = (
        "variable_versions",
        "variable_traces",
        "loop_variables",
        "break_collections",
        "continue_collections",
        "return_collections",
        "exception_collections",
        "outline_functions",
        "very_trusted_module_variables",
        "module_usage_attempts",
        "distribution_names",
        "delayed_work",
    )

    def __init__(self, module, very_trusted_module_variables, old_collection):
        assert module.isCompiledPythonModule(), module

        CollectionStartPointMixin.__init__(self)

        TraceCollectionBase.__init__(
            self,
            owner=module,
            name=module.getFullName(),
            parent=None,
        )

        self.very_trusted_module_variables = very_trusted_module_variables

        # Attempts to use a module in this module.
        self.module_usage_attempts = OrderedSet()

        # Attempts to use a distribution in this module.
        self.distribution_names = OrderedDict()

        for module_variable in module.locals_scope.getLocalsRelevantVariables():
            self.initVariableModule(module_variable, old_collection)
            self.variable_escapable.add(module_variable)

        for temp_variable in module.getTempVariables(outline=None):
            self.initVariableUninitialized(temp_variable, old_collection)

        self.has_unescaped_variables = True

    def getVeryTrustedModuleVariables(self):
        return self.very_trusted_module_variables

    def updateVeryTrustedModuleVariables(self, very_trusted_module_variables):
        result = self.very_trusted_module_variables != very_trusted_module_variables

        self.very_trusted_module_variables = very_trusted_module_variables

        return result

    def getModuleUsageAttempts(self):
        return self.module_usage_attempts

    def onModuleUsageAttempts(self, module_usage_attempts):
        self.module_usage_attempts.update(module_usage_attempts)

    def onModuleUsageAttempt(self, module_usage_attempt):
        self.module_usage_attempts.add(module_usage_attempt)

    def getUsedDistributions(self):
        return self.distribution_names

    def onDistributionUsed(self, distribution_name, node, success):
        inclusion_logger.info_to_file_only(
            "Cannot find distribution '%s' at '%s', expect potential run time problem, unless this is unused code."
            % (distribution_name, node.source_ref.getAsString())
        )

        self.distribution_names[distribution_name] = success


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.gnu.org/licenses/agpl.txt
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Format related nodes format/bin/oct/hex/ascii.

These will most often be used for outputs, and the hope is, the type prediction or the
result prediction will help to be smarter, but generally these should not be that much
about performance critical.

"""

from nuitka.PythonVersions import python_version
from nuitka.specs import BuiltinParameterSpecs

from .ChildrenHavingMixins import (
    ChildrenHavingValueFormatSpecOptionalAutoNoneEmptyStrMixin,
)
from .ExpressionBases import ExpressionBase, ExpressionBuiltinSingleArgBase
from .ExpressionShapeMixins import (
    ExpressionIntOrLongExactMixin,
    ExpressionStrOrUnicodeExactMixin,
    ExpressionStrShapeExactMixin,
)
from .NodeMakingHelpers import makeStatementExpressionOnlyReplacementNode


class ExpressionBuiltinFormat(
    ExpressionStrOrUnicodeExactMixin,
    ChildrenHavingValueFormatSpecOptionalAutoNoneEmptyStrMixin,
    ExpressionBase,
):
    kind = "EXPRESSION_BUILTIN_FORMAT"

    named_children = ("value", "format_spec|auto_none_empty_str")

    def __init__(self, value, format_spec, source_ref):
        ChildrenHavingValueFormatSpecOptionalAutoNoneEmptyStrMixin.__init__(
            self,
            value=value,
            format_spec=format_spec,
        )

        ExpressionBase.__init__(self, source_ref)

    def computeExpression(self, trace_collection):
        # TODO: Can use the format built-in on compile time constants at least.

        value = self.subnode_value
        format_spec = self.subnode_format_spec

        # TODO: Provide "__format__" slot based handling.

        # Any code could be run, note that.
        trace_collection.onControlFlowEscape(self)

        # Any exception may be raised.
        trace_collection.onExceptionRaiseExit(BaseException)

        # Strings format themselves as what they are.
        if format_spec is None:
            if value.hasShapeStrOrUnicodeExact():
                return (
                    value,
                    "new_expression",
                    """\
Removed useless 'format' on '%s' value."""
                    % value.getTypeShape().getTypeName(),
                )

        return self, None, None


class ExpressionBuiltinAscii(
    ExpressionStrShapeExactMixin, ExpressionBuiltinSingleArgBase
):
    kind = "EXPRESSION_BUILTIN_ASCII"

    if python_version >= 0x300:
        builtin_spec = BuiltinParameterSpecs.builtin_ascii_spec


class ExpressionBuiltinBin(
    ExpressionStrShapeExactMixin, ExpressionBuiltinSingleArgBase
):
    kind = "EXPRESSION_BUILTIN_BIN"

    builtin_spec = BuiltinParameterSpecs.builtin_bin_spec


class ExpressionBuiltinOct(
    ExpressionStrShapeExactMixin, ExpressionBuiltinSingleArgBase
):
    kind = "EXPRESSION_BUILTIN_OCT"

    builtin_spec = BuiltinParameterSpecs.builtin_oct_spec


class ExpressionBuiltinHex(
    ExpressionStrShapeExactMixin, ExpressionBuiltinSingleArgBase
):
    kind = "EXPRESSION_BUILTIN_HEX"

    builtin_spec = BuiltinParameterSpecs.builtin_hex_spec


class ExpressionBuiltinId(
    ExpressionIntOrLongExactMixin, ExpressionBuiltinSingleArgBase
):
    kind = "EXPRESSION_BUILTIN_ID"

    builtin_spec = BuiltinParameterSpecs.builtin_id_spec

    def computeExpression(self, trace_collection):
        # Note: Quite impossible to predict the pointer value or anything, but
        # we know the result will be a long.
        return self, None, None

    def mayRaiseException(self, exception_type):
        return self.subnode_value.mayRaiseException(exception_type)

    def getIntValue(self):
        return self

    # TODO: Make use SideEffectsFromChildrenMixin in form of a SideEffectsFromChildMixin
    def computeExpressionDrop(self, statement, trace_collection):
        result = makeStatementExpressionOnlyReplacementNode(
            expression=self.subnode_value, node=self
        )

        del self.parent

        return (
            result,
            "new_statements",
            """\
Removed id taking for unused result.""",
        )

    def mayHaveSideEffects(self):
        return self.subnode_value.mayHaveSideEffects()

    def extractSideEffects(self):
        return self.subnode_value.extractSideEffects()


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.gnu.org/licenses/agpl.txt
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Outputs to the user.

Printing with intends or plain, mostly a compensation for the print strangeness.

We want to avoid "from __future__ import print_function" in every file out
there, which makes adding another debug print rather tedious. This should
cover all calls/uses of "print" we have to do, and the make it easy to simply
to "print for_debug" without much hassle (braces).

"""

from __future__ import print_function

import inspect
import os
import struct
import sys
import textwrap
import traceback

from nuitka.__past__ import (  # pylint: disable=redefined-builtin
    BrokenPipeError,
    raw_input,
)
from nuitka.utils.Utils import isWin32Windows

# Written by Options module.
is_quiet = False

# We have to interact with displayed progress bars when doing out trace outputs.
progress = None


def setQuiet():
    # singleton, pylint: disable=global-statement
    global is_quiet
    is_quiet = True


def isQuiet():
    return is_quiet


def printIndented(level, *what):
    print("    " * level, *what)


def printSeparator(level=0):
    print("    " * level, "*" * 10)


def printLine(*what, **kwargs):
    is_atty = sys.stdout.isatty()

    if progress and is_atty:
        progress.close()

    print(*what, **kwargs)


def printError(message):
    my_print(message, file=sys.stderr)


def flushStandardOutputs():
    sys.stdout.flush()
    sys.stderr.flush()


def getEnableStyleCode(style):
    style = _aliasStyle(style)

    if style == "pink":
        style = "\033[95m"
    elif style == "blue":
        style = "\033[94m"
    elif style == "green":
        style = "\033[92m"
    elif style == "yellow":
        style = "\033[93m"
    elif style == "red":
        style = "\033[91m"
    elif style == "bold":
        style = "\033[1m"
    elif style == "underline":
        style = "\033[4m"
    else:
        style = None

    return style


def wrapWithStyles(value, styles):
    assert type(styles) is not str
    result = (
        "".join(getEnableStyleCode(style) for style in styles)
        + value
        + getDisableStylesCode()
    )

    assert type(result) is str
    return result


_enabled_ansi = False


def _enableAnsi():
    # singleton, pylint: disable=global-statement
    global _enabled_ansi
    if not _enabled_ansi:
        # Only necessary on Windows, as a side effect of this, ANSI colors get enabled
        # for the terminal and never deactivated, so we are free to use them after
        # this.
        if os.name == "nt":
            os.system("")

        _enabled_ansi = True


def hasTerminalLinkSupport():
    if not sys.stdout.isatty():
        return False

    terminal_name = os.getenv("TERM_PROGRAM")

    if terminal_name is not None:
        if terminal_name == "vscode":
            return True

    # Windows terminal, in some cases, but not on Windows 11 unfortunately there
    # is no difference between CMD with its own terminal and CMD in Windows
    # Terminal, not even parent processes.
    if isWin32Windows():
        if "WT_SESSION" in os.environ or "WT_PROFILE_ID" in os.environ:
            return True

    # Powershell, should be used in new terminals only. Also hard to detect,
    # although here we could search parent process names to determine it.
    if isWin32Windows():
        if "PSModulePath" in os.environ:
            return True

    return False


def formatTerminalLink(text, link):
    if link is not None and hasTerminalLinkSupport():
        return "\033]8;;%s\033\\%s\033]8;;\033\\" % (link, text)
    else:
        return text


def getDisableStylesCode():
    return "\033[0m"


def _getIoctlGWINSZ(fd):
    # On non-Windows, this might work, # spell-checker: ignore GWINSZ,TIOCGWINSZ
    try:
        # pylint: disable=I0021,import-error
        import fcntl
        import termios

        return struct.unpack("hh", fcntl.ioctl(fd, termios.TIOCGWINSZ, "1234"))
    except BaseException:  # Catch all the things, pylint: disable=broad-except
        return None


def _getTerminalSizeWin32():
    try:
        from ctypes import create_string_buffer, windll

        # handles are -10 to -12
        h = windll.kernel32.GetStdHandle(-12)
        buffer = create_string_buffer(22)

        res = windll.kernel32.GetConsoleScreenBufferInfo(h, buffer)
        if res:
            # TODO: Could also define the C structure
            (
                _,
                _,
                _,
                _,
                _,
                left,
                top,
                right,
                bottom,
                _,
                _,
            ) = struct.unpack("hhhhHhhhhhh", buffer.raw)

            columns = right - left + 1
            lines = bottom - top + 1

            return lines, columns
    except BaseException:  # Catch all the things, pylint: disable=broad-except
        return None


def _getTerminalSize():
    # For Python2, we get to do it ourselves.

    if str is bytes:
        if isWin32Windows():
            columns = _getTerminalSizeWin32()
        else:
            # Try all file handles.
            columns = _getIoctlGWINSZ(0) or _getIoctlGWINSZ(1) or _getIoctlGWINSZ(2)

        if columns:
            return columns[1]

        try:
            return int(os.getenv("COLUMNS", "1000"))
        except ValueError:
            return 1000
    else:
        try:
            return os.get_terminal_size()[0]
        except OSError:
            return 1000


def _aliasStyle(style):
    """Map named styles to concrete colors."""

    # return driven, pylint: disable=too-many-return-statements

    if style == "test-prepare":
        return "pink"
    if style == "test-execution":
        return "green"
    if style == "test-progress":
        return "blue"
    if style == "test-debug":
        return "bold"
    if style == "link":
        return "blue"
    if style == "warning":
        return "yellow"
    if style == "scons-unexpected":
        return "yellow"
    if style == "error":
        return "red"
    else:
        return style


def _my_print(is_atty, args, kwargs):
    file_output = kwargs.get("file", sys.stdout)

    try:
        _my_print2(is_atty, args, kwargs)

        # Flush the output.
        file_output.flush()
    except BrokenPipeError:
        pass


def _my_print2(is_atty, args, kwargs):
    if "style" in kwargs:
        style = kwargs["style"]
        del kwargs["style"]

        if "end" in kwargs:
            end = kwargs["end"]
            del kwargs["end"]
        else:
            end = "\n"

        if style is not None and is_atty:
            enable_style = getEnableStyleCode(style)

            if enable_style is None:
                raise ValueError(
                    "%r is an invalid value for keyword argument style" % style
                )

            _enableAnsi()

            print(enable_style, end="", **kwargs)

        if style is not None and is_atty:
            print(*args, end="", **kwargs)
            print(getDisableStylesCode(), end=end, **kwargs)
        else:
            print(*args, end=end, **kwargs)
    else:
        print(*args, **kwargs)


def canUseColor(file_output):
    """Can use color with the given file handle.

    This function is intended for globally disabling the feature
    potentially, but for that we would need to use it everywhere
    first.
    """
    return file_output.isatty()


def my_print(*args, **kwargs):
    """Make sure we flush after every print.

    Not even the "-u" option does more than that and this is easy enough.

    Use kwarg style=[option] to print in a style listed below
    """

    file_output = kwargs.get("file", sys.stdout)

    is_atty = file_output.isatty()

    if progress and is_atty:
        with progress.withExternalWritingPause():
            _my_print(is_atty, args, kwargs)
    else:
        _my_print(is_atty, args, kwargs)


class ReportingSystemExit(SystemExit):
    """Our own system exit, after which a report should be written."""

    def __init__(self, exit_code, exit_message):
        SystemExit.__init__(self, exit_code)

        self.exit_message = exit_message


class OurLogger(object):
    def __init__(self, name, quiet=False, base_style=None):
        self.name = name
        self.base_style = base_style
        self.is_quiet = quiet

    def my_print(self, message, **kwargs):
        # For overload, pylint: disable=no-self-use
        my_print(message, **kwargs)

    @staticmethod
    def _warnMnemonic(mnemonic, style, output_function):
        if mnemonic.startswith("http"):
            url = mnemonic
            extra_prefix = ""
        else:
            url = "https://nuitka.net/info/%s.html" % mnemonic
            extra_prefix = "Complex topic! "

        output_function(
            """    %sMore information can be found at %s%s"""
            % (extra_prefix, getEnableStyleCode("link"), url),
            style=style,
        )

    def _printFormatted(self, prefix, leader, message, style, keep_format, file_handle):
        style = style or self.base_style

        if sys.stderr.isatty() and not keep_format:
            width = _getTerminalSize() or 10000
        else:
            width = 10000

        if leader is not None:
            message_prefix = prefix + "  "
        else:
            message_prefix = prefix

        if width - len(message_prefix) < 20:
            width = 80

        formatted_message = textwrap.fill(
            message,
            width=width,
            initial_indent=message_prefix,
            subsequent_indent=message_prefix,
            break_on_hyphens=False,
            break_long_words=False,
            expand_tabs=False,
            replace_whitespace=False,
        )

        formatted_message = _removeNotBreakingSpaces(formatted_message)

        if leader is not None:
            formatted_message = prefix + leader + "\n" + formatted_message

        self.my_print(formatted_message, style=style, file=file_handle)

    def warning(
        self, message, style="yellow", mnemonic=None, keep_format=False, leader=None
    ):
        if mnemonic is not None:
            from .options.Options import shallDisplayWarningMnemonic

            if not shallDisplayWarningMnemonic(mnemonic):
                return

        if self.name:
            prefix = "%s:WARNING: " % self.name
        else:
            prefix = "WARNING: "

        self._printFormatted(
            prefix=prefix,
            leader=leader,
            message=message,
            style=style,
            keep_format=keep_format,
            file_handle=sys.stderr,
        )

        if mnemonic is not None:
            self._warnMnemonic(mnemonic, style=style, output_function=self.warning)

    def sysexit(
        self, message="", style=None, mnemonic=None, exit_code=1, reporting=False
    ):
        from nuitka.Progress import closeProgressBar

        closeProgressBar()

        if exit_code != 0 and style is None:
            style = "red"

        if message:
            if exit_code != 0:
                self.my_print(
                    "FATAL: %s" % message,
                    style=style,
                    file=sys.stderr,
                )
            else:
                self.my_print(
                    message,
                    style=style,
                    file=sys.stderr,
                )

        if mnemonic is not None:
            self._warnMnemonic(mnemonic, style=style, output_function=self.warning)

        if reporting:
            raise ReportingSystemExit(exit_code=exit_code, exit_message=message)

        sys.exit(exit_code)

    def sysexit_exception(self, message, exception, exit_code=1):
        self.my_print("FATAL: %s" % message, style="red", file=sys.stderr)

        traceback.print_exc()
        self.sysexit("FATAL:" + repr(exception), exit_code=exit_code, reporting=True)

    def isQuiet(self):
        return is_quiet or self.is_quiet

    def info(
        self,
        message,
        style=None,
        mnemonic=None,
        prefix=None,
        keep_format=False,
        leader=None,
    ):
        if not self.isQuiet():
            self._printFormatted(
                prefix=(
                    "%s: " % self.name
                    if prefix is None
                    else "%s:%s: " % (self.name, prefix)
                ),
                leader=leader,
                message=message,
                style=style,
                keep_format=keep_format,
                file_handle=sys.stderr,
            )

            if mnemonic is not None:
                self._warnMnemonic(mnemonic, style=style, output_function=self.info)


class FileLogger(OurLogger):
    def __init__(self, name, quiet=False, base_style=None, file_handle=None):
        OurLogger.__init__(self, name=name, quiet=quiet, base_style=base_style)

        self.file_handle = file_handle

    def __del__(self):
        if self.file_handle is not None:
            self.file_handle.close()

    def my_print(self, message, **kwargs):
        if "file" not in kwargs:
            kwargs["file"] = self.file_handle or sys.stdout

        my_print(message, **kwargs)
        kwargs["file"].flush()

    def setFileHandle(self, file_handle):
        self.file_handle = file_handle

    def isFileOutput(self):
        return self.file_handle is not None

    def info(
        self,
        message,
        style=None,
        mnemonic=None,
        prefix=None,
        keep_format=False,
        leader=None,
    ):
        if not self.isQuiet() or self.file_handle:
            self._printFormatted(
                prefix=(
                    "%s: " % self.name
                    if prefix is None
                    else "%s:%s: " % (self.name, prefix)
                ),
                leader=leader,
                message=message,
                style=style,
                keep_format=keep_format,
                file_handle=self.file_handle or sys.stderr,
            )

            if mnemonic is not None:
                self._warnMnemonic(mnemonic, style=style, output_function=self.info)

    def debug(self, message, style=None):
        if self.file_handle:
            message = "%s:DEBUG: %s" % (self.name, message)

            style = style or self.base_style
            self.my_print(message, style=style)

    def info_to_file_only(self, message, style=None):
        if self.file_handle:
            self.info(message, style=style)

    def info_if_file(self, message, other_logger, style=None):
        if self.file_handle:
            self.info(message, style=style)
        else:
            other_logger.info(message, style=style)


class OurSconsLogger(OurLogger):
    def sysexit(
        self,
        message,
        style=None,
        mnemonic=None,
        exit_code=1,
        reporting=False,
        env=None,
    ):
        if env is not None:
            from nuitka.utils.FileOperations import getNormalizedPathJoin
            from nuitka.utils.Json import writeJsonToFilename

            writeJsonToFilename(
                getNormalizedPathJoin(env.source_dir, "scons-error.json"),
                {
                    "message": message,
                    "mnemonic": mnemonic,
                    "exit_code": exit_code,
                    "reporting": reporting,
                },
            )

            sys.exit(27)
        else:
            OurLogger.sysexit(
                self,
                message,
                style=style,
                mnemonic=mnemonic,
                exit_code=exit_code,
                reporting=reporting,
            )


general = OurLogger("Nuitka")
plugins_logger = OurLogger("Nuitka-Plugins")
recursion_logger = OurLogger("Nuitka-Inclusion")
progress_logger = OurLogger("Nuitka-Progress", quiet=True)
memory_logger = OurLogger("Nuitka-Memory")
metadata_logger = OurLogger("Nuitka-Metadata")
optimization_logger = FileLogger("Nuitka-Optimization")
pgo_logger = FileLogger("Nuitka-PGO")
code_generation_logger = OurLogger("Nuitka-CodeGen")
inclusion_logger = FileLogger("Nuitka-Inclusion")
scons_logger = OurSconsLogger("Nuitka-Scons")
scons_details_logger = OurLogger("Nuitka-Scons")
postprocessing_logger = OurLogger("Nuitka-Postprocessing")
options_logger = OurLogger("Nuitka-Options")
unusual_logger = OurLogger("Nuitka-Unusual")
data_composer_logger = OurLogger("Nuitka-DataComposer")
onefile_logger = OurLogger("Nuitka-Onefile")
tools_logger = OurLogger("Nuitka-Tools")
wheel_logger = OurLogger("Nuitka-Wheel", base_style="blue")
cache_logger = OurLogger("Nuitka-Cache")
reports_logger = OurLogger("Nuitka-Reports")
offsets_logger = OurLogger("Nuitka-Offsets")


def queryUser(question, choices, default, default_non_interactive):
    assert default in choices, (default, choices)
    assert default_non_interactive in choices, (default, choices)

    prompt = "%s? %s : " % (
        question,
        "/".join(
            "[%s]" % choice.title() if choice == default else choice.title()
            for choice in choices
        ),
    )

    # Integrates with progress bar by closing it.
    printLine(prompt, end="")
    flushStandardOutputs()

    try:
        reply = raw_input() or default
    except EOFError:
        reply = default_non_interactive
        printLine("%s (default non-interactive)" % (reply))

    reply = reply.lower()

    if reply == "y":
        reply = "yes"

    return reply


_non_breaking_space = chr(1)


def doNotBreakSpaces(*args):
    if len(args) == 1 and inspect.isgenerator(args[0]):
        return doNotBreakSpaces(*tuple(args[0]))

    return tuple(element.replace(" ", _non_breaking_space) for element in args)


class LineBufferedStreamRedirector(object):
    def __init__(self, stream):
        self.original_stream = stream
        self.buffer = ""

    def write(self, message):
        if not message:
            return

        self.buffer += message

        if "\n" in self.buffer:
            lines = self.buffer.split("\n")
            # All lines except the last one are complete.
            # The last one is the start of the new line (remainder).
            if lines[-1] == "":
                # Ends with a newline
                to_print = self.buffer
                self.buffer = ""
            else:
                to_print = "\n".join(lines[:-1]) + "\n"
                self.buffer = lines[-1]

            if progress and self.original_stream.isatty():
                with progress.withExternalWritingPause():
                    self.original_stream.write(to_print)
                    self.original_stream.flush()
            else:
                self.original_stream.write(to_print)
                self.original_stream.flush()

    def flush(self):
        # We don't flush the buffer, as that would cause partial lines to be
        # printed, which then get overwritten by the progress bar. We wait
        # for the newline to print it.
        self.original_stream.flush()

    def __del__(self):
        if self.buffer:
            try:
                self.original_stream.write(self.buffer)
                self.original_stream.flush()
            except Exception:  # pylint: disable=broad-except
                pass

    def isatty(self):
        return self.original_stream.isatty()


def _removeNotBreakingSpaces(message):
    #    return message
    return message.replace(_non_breaking_space, " ")


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.gnu.org/licenses/agpl.txt
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""
This module handles the extraction of build configuration from Poetry projects.
"""

import os
import sys

from nuitka.importing.Importing import addMainScriptDirectory
from nuitka.utils.Execution import executeProcess
from nuitka.utils.FileOperations import withTemporaryDirectory
from nuitka.utils.Json import loadJsonFromFilename

from .BuildPackageCommon import reportBuildError, setProjectName


def getPoetryBuildConfiguration(logger):
    """
    Get the build configuration from a Poetry project.
    """
    # This is "poetry" handling.
    with withTemporaryDirectory("nuitka-project-dump") as temp_dir:
        dump_filename = os.path.join(temp_dir, "build_config.json")

        # Locate the script "misc/extract-poetry-config.py" relative to this module.
        # We assume this module is in "nuitka/options/Poetry.py"
        script_filename = os.path.join(
            os.path.dirname(__file__),
            "..",
            "tools",
            "general",
            "extract_poetry_config",
        )

        command = (
            sys.executable,
            script_filename,
            dump_filename,
        )

        stdout, stderr, exit_code = executeProcess(
            command,
            stdin=False,
        )

        if exit_code != 0:
            reportBuildError(logger, "poetry", command, stdout, stderr)

        config = loadJsonFromFilename(dump_filename)

        package_dir = config.get("package_dir")
        if package_dir and "" in package_dir:
            addMainScriptDirectory(os.path.join(os.getcwd(), package_dir.get("")))
        else:
            addMainScriptDirectory(os.getcwd())

    setProjectName(config.get("project_name"))

    return config.get("arguments", [])


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.gnu.org/licenses/agpl.txt
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



ord_names = {
    2: b'SysAllocString',
    3: b'SysReAllocString',
    4: b'SysAllocStringLen',
    5: b'SysReAllocStringLen',
    6: b'SysFreeString',
    7: b'SysStringLen',
    8: b'VariantInit',
    9: b'VariantClear',
    10: b'VariantCopy',
    11: b'VariantCopyInd',
    12: b'VariantChangeType',
    13: b'VariantTimeToDosDateTime',
    14: b'DosDateTimeToVariantTime',
    15: b'SafeArrayCreate',
    16: b'SafeArrayDestroy',
    17: b'SafeArrayGetDim',
    18: b'SafeArrayGetElemsize',
    19: b'SafeArrayGetUBound',
    20: b'SafeArrayGetLBound',
    21: b'SafeArrayLock',
    22: b'SafeArrayUnlock',
    23: b'SafeArrayAccessData',
    24: b'SafeArrayUnaccessData',
    25: b'SafeArrayGetElement',
    26: b'SafeArrayPutElement',
    27: b'SafeArrayCopy',
    28: b'DispGetParam',
    29: b'DispGetIDsOfNames',
    30: b'DispInvoke',
    31: b'CreateDispTypeInfo',
    32: b'CreateStdDispatch',
    33: b'RegisterActiveObject',
    34: b'RevokeActiveObject',
    35: b'GetActiveObject',
    36: b'SafeArrayAllocDescriptor',
    37: b'SafeArrayAllocData',
    38: b'SafeArrayDestroyDescriptor',
    39: b'SafeArrayDestroyData',
    40: b'SafeArrayRedim',
    41: b'SafeArrayAllocDescriptorEx',
    42: b'SafeArrayCreateEx',
    43: b'SafeArrayCreateVectorEx',
    44: b'SafeArraySetRecordInfo',
    45: b'SafeArrayGetRecordInfo',
    46: b'VarParseNumFromStr',
    47: b'VarNumFromParseNum',
    48: b'VarI2FromUI1',
    49: b'VarI2FromI4',
    50: b'VarI2FromR4',
    51: b'VarI2FromR8',
    52: b'VarI2FromCy',
    53: b'VarI2FromDate',
    54: b'VarI2FromStr',
    55: b'VarI2FromDisp',
    56: b'VarI2FromBool',
    57: b'SafeArraySetIID',
    58: b'VarI4FromUI1',
    59: b'VarI4FromI2',
    60: b'VarI4FromR4',
    61: b'VarI4FromR8',
    62: b'VarI4FromCy',
    63: b'VarI4FromDate',
    64: b'VarI4FromStr',
    65: b'VarI4FromDisp',
    66: b'VarI4FromBool',
    67: b'SafeArrayGetIID',
    68: b'VarR4FromUI1',
    69: b'VarR4FromI2',
    70: b'VarR4FromI4',
    71: b'VarR4FromR8',
    72: b'VarR4FromCy',
    73: b'VarR4FromDate',
    74: b'VarR4FromStr',
    75: b'VarR4FromDisp',
    76: b'VarR4FromBool',
    77: b'SafeArrayGetVartype',
    78: b'VarR8FromUI1',
    79: b'VarR8FromI2',
    80: b'VarR8FromI4',
    81: b'VarR8FromR4',
    82: b'VarR8FromCy',
    83: b'VarR8FromDate',
    84: b'VarR8FromStr',
    85: b'VarR8FromDisp',
    86: b'VarR8FromBool',
    87: b'VarFormat',
    88: b'VarDateFromUI1',
    89: b'VarDateFromI2',
    90: b'VarDateFromI4',
    91: b'VarDateFromR4',
    92: b'VarDateFromR8',
    93: b'VarDateFromCy',
    94: b'VarDateFromStr',
    95: b'VarDateFromDisp',
    96: b'VarDateFromBool',
    97: b'VarFormatDateTime',
    98: b'VarCyFromUI1',
    99: b'VarCyFromI2',
    100: b'VarCyFromI4',
    101: b'VarCyFromR4',
    102: b'VarCyFromR8',
    103: b'VarCyFromDate',
    104: b'VarCyFromStr',
    105: b'VarCyFromDisp',
    106: b'VarCyFromBool',
    107: b'VarFormatNumber',
    108: b'VarBstrFromUI1',
    109: b'VarBstrFromI2',
    110: b'VarBstrFromI4',
    111: b'VarBstrFromR4',
    112: b'VarBstrFromR8',
    113: b'VarBstrFromCy',
    114: b'VarBstrFromDate',
    115: b'VarBstrFromDisp',
    116: b'VarBstrFromBool',
    117: b'VarFormatPercent',
    118: b'VarBoolFromUI1',
    119: b'VarBoolFromI2',
    120: b'VarBoolFromI4',
    121: b'VarBoolFromR4',
    122: b'VarBoolFromR8',
    123: b'VarBoolFromDate',
    124: b'VarBoolFromCy',
    125: b'VarBoolFromStr',
    126: b'VarBoolFromDisp',
    127: b'VarFormatCurrency',
    128: b'VarWeekdayName',
    129: b'VarMonthName',
    130: b'VarUI1FromI2',
    131: b'VarUI1FromI4',
    132: b'VarUI1FromR4',
    133: b'VarUI1FromR8',
    134: b'VarUI1FromCy',
    135: b'VarUI1FromDate',
    136: b'VarUI1FromStr',
    137: b'VarUI1FromDisp',
    138: b'VarUI1FromBool',
    139: b'VarFormatFromTokens',
    140: b'VarTokenizeFormatString',
    141: b'VarAdd',
    142: b'VarAnd',
    143: b'VarDiv',
    144: b'DllCanUnloadNow',
    145: b'DllGetClassObject',
    146: b'DispCallFunc',
    147: b'VariantChangeTypeEx',
    148: b'SafeArrayPtrOfIndex',
    149: b'SysStringByteLen',
    150: b'SysAllocStringByteLen',
    151: b'DllRegisterServer',
    152: b'VarEqv',
    153: b'VarIdiv',
    154: b'VarImp',
    155: b'VarMod',
    156: b'VarMul',
    157: b'VarOr',
    158: b'VarPow',
    159: b'VarSub',
    160: b'CreateTypeLib',
    161: b'LoadTypeLib',
    162: b'LoadRegTypeLib',
    163: b'RegisterTypeLib',
    164: b'QueryPathOfRegTypeLib',
    165: b'LHashValOfNameSys',
    166: b'LHashValOfNameSysA',
    167: b'VarXor',
    168: b'VarAbs',
    169: b'VarFix',
    170: b'OaBuildVersion',
    171: b'ClearCustData',
    172: b'VarInt',
    173: b'VarNeg',
    174: b'VarNot',
    175: b'VarRound',
    176: b'VarCmp',
    177: b'VarDecAdd',
    178: b'VarDecDiv',
    179: b'VarDecMul',
    180: b'CreateTypeLib2',
    181: b'VarDecSub',
    182: b'VarDecAbs',
    183: b'LoadTypeLibEx',
    184: b'SystemTimeToVariantTime',
    185: b'VariantTimeToSystemTime',
    186: b'UnRegisterTypeLib',
    187: b'VarDecFix',
    188: b'VarDecInt',
    189: b'VarDecNeg',
    190: b'VarDecFromUI1',
    191: b'VarDecFromI2',
    192: b'VarDecFromI4',
    193: b'VarDecFromR4',
    194: b'VarDecFromR8',
    195: b'VarDecFromDate',
    196: b'VarDecFromCy',
    197: b'VarDecFromStr',
    198: b'VarDecFromDisp',
    199: b'VarDecFromBool',
    200: b'GetErrorInfo',
    201: b'SetErrorInfo',
    202: b'CreateErrorInfo',
    203: b'VarDecRound',
    204: b'VarDecCmp',
    205: b'VarI2FromI1',
    206: b'VarI2FromUI2',
    207: b'VarI2FromUI4',
    208: b'VarI2FromDec',
    209: b'VarI4FromI1',
    210: b'VarI4FromUI2',
    211: b'VarI4FromUI4',
    212: b'VarI4FromDec',
    213: b'VarR4FromI1',
    214: b'VarR4FromUI2',
    215: b'VarR4FromUI4',
    216: b'VarR4FromDec',
    217: b'VarR8FromI1',
    218: b'VarR8FromUI2',
    219: b'VarR8FromUI4',
    220: b'VarR8FromDec',
    221: b'VarDateFromI1',
    222: b'VarDateFromUI2',
    223: b'VarDateFromUI4',
    224: b'VarDateFromDec',
    225: b'VarCyFromI1',
    226: b'VarCyFromUI2',
    227: b'VarCyFromUI4',
    228: b'VarCyFromDec',
    229: b'VarBstrFromI1',
    230: b'VarBstrFromUI2',
    231: b'VarBstrFromUI4',
    232: b'VarBstrFromDec',
    233: b'VarBoolFromI1',
    234: b'VarBoolFromUI2',
    235: b'VarBoolFromUI4',
    236: b'VarBoolFromDec',
    237: b'VarUI1FromI1',
    238: b'VarUI1FromUI2',
    239: b'VarUI1FromUI4',
    240: b'VarUI1FromDec',
    241: b'VarDecFromI1',
    242: b'VarDecFromUI2',
    243: b'VarDecFromUI4',
    244: b'VarI1FromUI1',
    245: b'VarI1FromI2',
    246: b'VarI1FromI4',
    247: b'VarI1FromR4',
    248: b'VarI1FromR8',
    249: b'VarI1FromDate',
    250: b'VarI1FromCy',
    251: b'VarI1FromStr',
    252: b'VarI1FromDisp',
    253: b'VarI1FromBool',
    254: b'VarI1FromUI2',
    255: b'VarI1FromUI4',
    256: b'VarI1FromDec',
    257: b'VarUI2FromUI1',
    258: b'VarUI2FromI2',
    259: b'VarUI2FromI4',
    260: b'VarUI2FromR4',
    261: b'VarUI2FromR8',
    262: b'VarUI2FromDate',
    263: b'VarUI2FromCy',
    264: b'VarUI2FromStr',
    265: b'VarUI2FromDisp',
    266: b'VarUI2FromBool',
    267: b'VarUI2FromI1',
    268: b'VarUI2FromUI4',
    269: b'VarUI2FromDec',
    270: b'VarUI4FromUI1',
    271: b'VarUI4FromI2',
    272: b'VarUI4FromI4',
    273: b'VarUI4FromR4',
    274: b'VarUI4FromR8',
    275: b'VarUI4FromDate',
    276: b'VarUI4FromCy',
    277: b'VarUI4FromStr',
    278: b'VarUI4FromDisp',
    279: b'VarUI4FromBool',
    280: b'VarUI4FromI1',
    281: b'VarUI4FromUI2',
    282: b'VarUI4FromDec',
    283: b'BSTR_UserSize',
    284: b'BSTR_UserMarshal',
    285: b'BSTR_UserUnmarshal',
    286: b'BSTR_UserFree',
    287: b'VARIANT_UserSize',
    288: b'VARIANT_UserMarshal',
    289: b'VARIANT_UserUnmarshal',
    290: b'VARIANT_UserFree',
    291: b'LPSAFEARRAY_UserSize',
    292: b'LPSAFEARRAY_UserMarshal',
    293: b'LPSAFEARRAY_UserUnmarshal',
    294: b'LPSAFEARRAY_UserFree',
    295: b'LPSAFEARRAY_Size',
    296: b'LPSAFEARRAY_Marshal',
    297: b'LPSAFEARRAY_Unmarshal',
    298: b'VarDecCmpR8',
    299: b'VarCyAdd',
    300: b'DllUnregisterServer',
    301: b'OACreateTypeLib2',
    303: b'VarCyMul',
    304: b'VarCyMulI4',
    305: b'VarCySub',
    306: b'VarCyAbs',
    307: b'VarCyFix',
    308: b'VarCyInt',
    309: b'VarCyNeg',
    310: b'VarCyRound',
    311: b'VarCyCmp',
    312: b'VarCyCmpR8',
    313: b'VarBstrCat',
    314: b'VarBstrCmp',
    315: b'VarR8Pow',
    316: b'VarR4CmpR8',
    317: b'VarR8Round',
    318: b'VarCat',
    319: b'VarDateFromUdateEx',
    322: b'GetRecordInfoFromGuids',
    323: b'GetRecordInfoFromTypeInfo',
    325: b'SetVarConversionLocaleSetting',
    326: b'GetVarConversionLocaleSetting',
    327: b'SetOaNoCache',
    329: b'VarCyMulI8',
    330: b'VarDateFromUdate',
    331: b'VarUdateFromDate',
    332: b'GetAltMonthNames',
    333: b'VarI8FromUI1',
    334: b'VarI8FromI2',
    335: b'VarI8FromR4',
    336: b'VarI8FromR8',
    337: b'VarI8FromCy',
    338: b'VarI8FromDate',
    339: b'VarI8FromStr',
    340: b'VarI8FromDisp',
    341: b'VarI8FromBool',
    342: b'VarI8FromI1',
    343: b'VarI8FromUI2',
    344: b'VarI8FromUI4',
    345: b'VarI8FromDec',
    346: b'VarI2FromI8',
    347: b'VarI2FromUI8',
    348: b'VarI4FromI8',
    349: b'VarI4FromUI8',
    360: b'VarR4FromI8',
    361: b'VarR4FromUI8',
    362: b'VarR8FromI8',
    363: b'VarR8FromUI8',
    364: b'VarDateFromI8',
    365: b'VarDateFromUI8',
    366: b'VarCyFromI8',
    367: b'VarCyFromUI8',
    368: b'VarBstrFromI8',
    369: b'VarBstrFromUI8',
    370: b'VarBoolFromI8',
    371: b'VarBoolFromUI8',
    372: b'VarUI1FromI8',
    373: b'VarUI1FromUI8',
    374: b'VarDecFromI8',
    375: b'VarDecFromUI8',
    376: b'VarI1FromI8',
    377: b'VarI1FromUI8',
    378: b'VarUI2FromI8',
    379: b'VarUI2FromUI8',
    401: b'OleLoadPictureEx',
    402: b'OleLoadPictureFileEx',
    411: b'SafeArrayCreateVector',
    412: b'SafeArrayCopyData',
    413: b'VectorFromBstr',
    414: b'BstrFromVector',
    415: b'OleIconToCursor',
    416: b'OleCreatePropertyFrameIndirect',
    417: b'OleCreatePropertyFrame',
    418: b'OleLoadPicture',
    419: b'OleCreatePictureIndirect',
    420: b'OleCreateFontIndirect',
    421: b'OleTranslateColor',
    422: b'OleLoadPictureFile',
    423: b'OleSavePictureFile',
    424: b'OleLoadPicturePath',
    425: b'VarUI4FromI8',
    426: b'VarUI4FromUI8',
    427: b'VarI8FromUI8',
    428: b'VarUI8FromI8',
    429: b'VarUI8FromUI1',
    430: b'VarUI8FromI2',
    431: b'VarUI8FromR4',
    432: b'VarUI8FromR8',
    433: b'VarUI8FromCy',
    434: b'VarUI8FromDate',
    435: b'VarUI8FromStr',
    436: b'VarUI8FromDisp',
    437: b'VarUI8FromBool',
    438: b'VarUI8FromI1',
    439: b'VarUI8FromUI2',
    440: b'VarUI8FromUI4',
    441: b'VarUI8FromDec',
    442: b'RegisterTypeLibForUser',
    443: b'UnRegisterTypeLibForUser',
}



#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Profiling for Nuitka and CPython.

This provides the capability of comparing performance results of Nuitka and
CPython relatively to one another.
"""

# Note: This is currently severely broken.

import runpy
import sys
import tempfile

from nuitka.Tracing import my_print


def _namelen(e):
    if e.startswith("py:"):
        return len(e.split(":")[1])
    else:
        return len(e)


def show(stats):
    p = stats.top_profile()
    if not p:
        my_print("no stats")
        return

    p.sort(key=lambda x: x[1], reverse=True)
    top = p[0][1]

    max_len = max(_namelen(e[0]) for e in p)

    my_print(" vmprof output:")
    my_print(" %:      name:" + " " * (max_len - 3) + "location:")

    for k, v in p:
        v = "%.1f%%" % (float(v) / top * 100)
        if v == "0.0%":
            v = "<0.1%"
        if k.startswith("py:"):
            _, func_name, lineno, filename = k.split(":", 3)
            lineno = int(lineno)
            my_print(
                " %s %s %s:%d"
                % (v.ljust(7), func_name.ljust(max_len + 1), filename, lineno)
            )
        else:
            my_print(" %s %s" % (v.ljust(7), k.ljust(max_len + 1)))


def main():
    import vmprof  # pylint: disable=I0021,import-error

    with tempfile.NamedTemporaryFile() as prof_file:
        vmprof.enable(prof_file.fileno(), 0.001)

        try:
            program = sys.argv[1]
            del sys.argv[1]

            #            sys.argv = [args.program] + args.args
            runpy.run_path(program, run_name="__main__")
        except (KeyboardInterrupt, SystemExit):
            pass

        vmprof.disable()

        stats = vmprof.read_profile(prof_file.name)

        show(stats)


if __name__ == "__main__":
    main()

#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.gnu.org/licenses/agpl.txt
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Nuitka yaml utility functions.

Because we want to work with Python2.6 or higher, we play a few tricks with
what library to use for what Python. We have an 2 inline copy of PyYAML, one
that still does 2.6 and one for newer Pythons.

Also we put loading for specific packages in here and a few helpers to work
with these config files.
"""

# Otherwise "Yaml" and "yaml" collide on case insensitive setups
from __future__ import absolute_import

import ast
import os
import pkgutil
import re
from posixpath import normpath

from nuitka.containers.OrderedDicts import OrderedDict
from nuitka.options.Options import getUserProvidedYamlFiles
from nuitka.Tracing import general

from .FileOperations import getFileContents
from .Hashing import HashCRC32
from .Importing import importFromInlineCopy
from .ModuleNames import checkModuleName
from .PrivatePipSpace import getPrivatePackage


def _isParsable(value):
    """Check if a value is parsable python code."""
    try:
        ast.parse(value)
    except (SyntaxError, IndentationError):
        return False
    else:
        return True


def _isNormalizedPosixPath(path):
    """Check if a path is a normalized POSIX path."""
    if "\\" in path:
        return False

    return path == normpath(path)


def _checkNotEmptyString(logger, filename, module_name, section, k, value):
    """Check if a string value is not empty and log error if it is."""
    if value == "":
        logger.info(
            """\
%s: %s config value of %s %s cannot be empty."""
            % (filename, module_name, section, k),
            keep_format=True,
        )
        return False

    return True


def _checkParsable(logger, filename, module_name, section, k, value):
    """Check if a value is parsable python code and log error if not."""
    if not _checkNotEmptyString(logger, filename, module_name, section, k, value):
        return False

    if not _isParsable(value):
        logger.info(
            """\
%s: %s config value of '%s' '%s' contains invalid syntax in value '%s'"""
            % (filename, module_name, section, k, value),
            keep_format=True,
        )
        return False

    return True


def _checkRegexp(logger, filename, module_name, section, k, regexp, replacement):
    """Check if a regexp value is valid and log error if not."""
    if not _checkNotEmptyString(logger, filename, module_name, section, k, regexp):
        return False

    try:
        re.sub(regexp, replacement, "", re.S)
    except re.error as e:
        logger.info(
            """\
%s: %s config value of '%s' '%s' contains invalid regexp \
syntax in value '%s' leading to error '%s'"""
            % (filename, module_name, section, regexp, replacement, e),
            keep_format=True,
        )
        return False

    return True


def _checkNormalizedPosixPath(logger, filename, module_name, section, k, value):
    """Check if a value is a normalized POSIX path and log error if not."""
    if not _isNormalizedPosixPath(value):
        logger.info(
            """\
%s: module '%s' config value of '%s' '%s' should be normalized posix \
path, with '/' style slashes not '%s'."""
            % (filename, module_name, section, k, value),
            keep_format=True,
        )
        return False

    return True


def checkSectionValues(logger, filename, module_name, section, value):
    """Check values of the YAML file."""
    # many checks of course, pylint: disable=too-many-branches,too-many-statements

    result = True

    if type(value) is dict:
        for k, v in value.items():
            if k == "description" and v != v.strip():
                logger.info(
                    """\
%s: %s config value of %s %s should not contain trailing or leading spaces"""
                    % (filename, module_name, section, k),
                    keep_format=True,
                )
                result = False

            if k in ("when", "append_result"):
                if not _checkParsable(logger, filename, module_name, section, k, v):
                    result = False

            if k in ("replacements", "global_replacements"):
                for m, d in v.items():
                    if not _checkNotEmptyString(
                        logger, filename, module_name, section, k, m
                    ):
                        result = False
                    elif not _checkParsable(
                        logger, filename, module_name, section, k, d
                    ):
                        result = False

            if k in ("replacements_re", "global_replacements_re"):
                for m, d in v.items():
                    if not _checkRegexp(
                        logger, filename, module_name, section, k, m, d
                    ):
                        result = False

            if k == "replacements_plain":
                for m, d in v.items():
                    if not _checkNotEmptyString(
                        logger, filename, module_name, section, k, m
                    ):
                        result = False

            if k in ("dest_path", "relative_path") and not _checkNormalizedPosixPath(
                logger, filename, module_name, section, k, v
            ):
                result = False

            if k in ("dirs", "raw_dirs", "empty_dirs"):
                for e in v:
                    if not _checkNormalizedPosixPath(
                        logger, filename, module_name, section, k, e
                    ):
                        result = False

            if k == "parameters":
                for item in v:
                    if "values" in item:
                        if not _checkParsable(
                            logger, filename, module_name, section, k, item["values"]
                        ):
                            result = False

            if k == "no-auto-follow":
                for m, d in v.items():
                    if d == "":
                        logger.info(
                            """\
%s: %s config value of %s %s should not use empty value for %s, use 'ignore' \
if you want no message."""
                            % (filename, module_name, section, k, m),
                            keep_format=True,
                        )
                        result = False

            if k == "declarations":
                for m, d in v.items():
                    if not _checkNotEmptyString(
                        logger, filename, module_name, section, k, m
                    ):
                        result = False

                    if not _checkParsable(logger, filename, module_name, section, k, d):
                        result = False

            if k == "append_plain":
                if not _checkParsable(logger, filename, module_name, section, k, v):
                    result = False

            if k == "setup_code":
                if type(v) is list:
                    if not _checkParsable(
                        logger, filename, module_name, section, k, "\n".join(v)
                    ):
                        result = False
                elif not _checkParsable(logger, filename, module_name, section, k, v):
                    result = False

            if k == "environment":
                for m, d in v.items():
                    if not _checkParsable(logger, filename, module_name, section, k, d):
                        result = False

            if not checkSectionValues(logger, filename, module_name, section, v):
                result = False
    elif type(value) in (list, tuple):
        for item in value:
            if not checkSectionValues(logger, filename, module_name, section, item):
                result = False

    return result


def getJsonschemaPackage(logger, assume_yes_for_downloads, reject_message):
    """Get jsonschema package from private pip space or globally."""
    return getPrivatePackage(
        logger=logger,
        package_name="jsonschema",
        module_name="jsonschema",
        package_version=None,
        submodule_names=("validators",),
        assume_yes_for_downloads=assume_yes_for_downloads,
        reject_message=reject_message,
    )


def getRuamelYamlPackage(logger, assume_yes_for_downloads):
    """Get ruamel.yaml package from private pip space or globally."""
    return getPrivatePackage(
        logger=logger,
        package_name="ruamel.yaml",
        module_name="ruamel.yaml",
        package_version=None,
        submodule_names=None,
        assume_yes_for_downloads=assume_yes_for_downloads,
        reject_message="Autoformat YAML needs ruamel.yaml.",
    )


def getYamllintPackage(logger, assume_yes_for_downloads, reject_message):
    """Get yamllint package from private pip space or globally."""
    return getPrivatePackage(
        logger=logger,
        package_name="yamllint",
        module_name="yamllint",
        package_version=None,
        submodule_names=("cli",),
        assume_yes_for_downloads=assume_yes_for_downloads,
        reject_message=reject_message,
    )


def getDeepDiffPackage(logger, assume_yes_for_downloads):
    """Get deepdiff package from private pip space or globally."""
    return getPrivatePackage(
        logger=logger,
        package_name="deepdiff",
        module_name="deepdiff",
        package_version=None,
        submodule_names=("diff",),
        assume_yes_for_downloads=assume_yes_for_downloads,
        reject_message="Autoformat YAML needs deepdiff.",
    )


def checkDataChecksums(file_data, data):
    # Checksum is valid, if the file contains a line with the checksum of the
    # file content without that line.

    result = []

    # We are parsing the file manually here, which is not great, but strict
    # enough for the auto-formatted files we have.
    for line in file_data.decode("utf8").splitlines():
        if not line.startswith("- module-name:"):
            continue

        parts = line.split(":", 2)

        module_name = parts[1]
        module_name = module_name.split("#", 2)[0]
        module_name = module_name.strip()
        module_name = module_name.strip("'\"")

        if "# checksum: " not in line:
            result.append(module_name)
            continue

        expected_checksum = line.split("# checksum: ")[1].strip()

        yaml_module_data = data.get(module_name)

        if getYamlDataHash(yaml_module_data) != expected_checksum:
            result.append(module_name)
            continue

    return result


def validateSchema(logger, name, data, assume_yes_for_downloads, reject_message):
    if not getJsonschemaPackage(
        logger,
        assume_yes_for_downloads=assume_yes_for_downloads,
        reject_message=reject_message,
    ):
        logger.warning("Cannot validate schema due to lack of 'jsonschema' package.")
        return

    # pylint: disable=I0021,import-error
    from jsonschema import validators

    schema_filename = getYamlPackageConfigurationSchemaFilename()

    if not os.path.exists(schema_filename):
        logger.sysexit("Cannot validate schema due to missing schema file.")

    import json

    schema = json.loads(getFileContents(schema_filename))

    validator = validators.Draft202012Validator(schema=schema)

    error_messages = []

    for error in validator.iter_errors(instance=data):
        try:
            module_name = repr(data[error.path[0]]["module-name"])
        except Exception:  # pylint: disable=broad-except
            module_name = "some"

        error_messages.append("For %s module: %s" % (module_name, error.message))

    if error_messages:
        logger.sysexit(
            "Error, invalid package configuration in '%s':\n%s"
            % (name, "\n".join(error_messages))
        )


class PackageConfigYaml(object):
    __slots__ = (
        "name",
        "data",
        "logger",
        "checked_sections",
    )

    def __init__(
        self, logger, name, file_data, assume_yes_for_downloads, check_checksums
    ):
        self.logger = logger
        self.name = name

        assert type(file_data) is bytes
        data = parseYaml(
            logger=logger,
            data=file_data,
            error_message="""\
Error, empty (or malformed?) user package configuration '%s' used."""
            % name,
        )

        assert type(data) is list, type(data)

        self._init(logger, data)

        if check_checksums:
            bad_checksum_modules = checkDataChecksums(file_data, self.data)

            if bad_checksum_modules:
                logger.info(
                    "Detected %d module(s) with mismatching checksum in '%s': %s"
                    % (len(bad_checksum_modules), name, ",".join(bad_checksum_modules))
                )

                validateSchema(
                    logger=logger,
                    name=name,
                    data=data,
                    assume_yes_for_downloads=assume_yes_for_downloads,
                    reject_message=None,
                )

    def _init(self, logger, data):
        self.data = OrderedDict()

        for item in data:
            module_name = item.get("module-name")

            if not module_name:
                return logger.sysexit(
                    "Error, invalid config in '%s' looks like an empty module name was given."
                    % (self.name)
                )

            if "/" in module_name:
                return logger.sysexit(
                    "Error, invalid module name in '%s' looks like a file path '%s'."
                    % (self.name, module_name)
                )

            if not checkModuleName(module_name):
                return logger.sysexit(
                    "Error, invalid module name in '%s' not valid '%s'."
                    % (self.name, module_name)
                )

            if module_name in self.data:
                return logger.sysexit(
                    "Duplicate module name '%s' encountered." % module_name
                )

            # Do not replicate module name in data.
            self.data[module_name] = item.copy()
            del self.data[module_name]["module-name"]

        self.checked_sections = set()

    def __repr__(self):
        return "<PackageConfigYaml %s>" % self.name

    def get(self, name, section):
        """Return a configs for that section."""
        result = self.data.get(name)

        if result is not None:
            result = result.get(section, ())
        else:
            result = ()

        if section in (
            "options",
            "variables",
            "constants",
        ) and type(
            result
        ) in (dict, OrderedDict):
            result = (result,)

        if type(result) is list:
            result = tuple(result)

        # Ensure result is a tuple; otherwise exit with an error
        if not isinstance(result, tuple):
            self.logger.sysexit(
                "Error, unexpected result type %s for %s module %s section %s."
                % (type(result), self.name, name, section)
            )

        if result and (name, section) not in self.checked_sections:
            checkSectionValues(self.logger, self.name, name, section, result)
            self.checked_sections.add((name, section))

        return result

    def keys(self):
        return self.data.keys()

    def items(self):
        return self.data.items()

    def update(self, other):
        # TODO: Full blown merging, including respecting an overload flag, where
        # a config replaces another one entirely, for now we expect to not
        # overlap and offer only merging of implicit-imports.
        for key, value in other.items():
            # assert key not in self.data, key
            if key in self.data:
                new_implicit_imports = value.get("implicit-imports", None)
                if new_implicit_imports:
                    value.pop("implicit-imports")
                    if self.data[key].get("implicit-imports", None) is None:
                        self.data[key]["implicit-imports"] = new_implicit_imports
                    else:
                        self.data[key]["implicit-imports"].extend(new_implicit_imports)
                if len(value) > 0:
                    general.sysexit(
                        "Error, duplicate config for module name '%s' encountered in '%s'."
                        % (key, self.name)
                    )
                else:
                    general.info("Merged implicit-imports for '%s'." % key)
            else:
                self.data[key] = value


def getYamlPackage():
    if not hasattr(getYamlPackage, "yaml"):
        try:
            import yaml

            getYamlPackage.yaml = yaml
        except ImportError:
            getYamlPackage.yaml = importFromInlineCopy(
                "yaml", must_exist=True, delete_module=True
            )

    return getYamlPackage.yaml


def parseYaml(logger, data, error_message):
    yaml = getYamlPackage()

    # Make sure dictionaries are ordered even before 3.6 in the result. We use
    # them for hashing in caching keys.
    class OrderedLoader(yaml.SafeLoader):
        pass

    def construct_mapping(loader, node):
        loader.flatten_mapping(node)

        return OrderedDict(loader.construct_pairs(node))

    OrderedLoader.add_constructor(
        yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG, construct_mapping
    )

    result = yaml.load(data, OrderedLoader)

    if not result:
        return logger.sysexit(error_message)

    return result


_yaml_cache = {}


def getYamlDataHash(data):
    result = HashCRC32()
    result.updateFromValues(data)

    return result.asHexDigest()


def parsePackageYaml(
    logger, package_name, filename, assume_yes_for_downloads, check_checksums
):
    key = package_name, filename

    if key not in _yaml_cache:
        if package_name is None:
            file_data = getFileContents(filename, mode="rb")
        else:
            file_data = pkgutil.get_data(package_name, filename)

        if file_data is None:
            raise IOError("Cannot find %s.%s" % (package_name, filename))

        _yaml_cache[key] = PackageConfigYaml(
            logger=logger,
            name=filename,
            file_data=file_data,
            assume_yes_for_downloads=assume_yes_for_downloads,
            check_checksums=check_checksums,
        )

    return _yaml_cache[key]


_package_config = None


def getYamlPackageConfiguration(logger, assume_yes_for_downloads, check_checksums):
    """Get Nuitka package configuration. Merged from multiple sources."""
    # Singleton, pylint: disable=global-statement
    global _package_config

    if logger is None:
        logger = general

    if _package_config is None:
        _package_config = parsePackageYaml(
            logger=logger,
            package_name="nuitka.plugins.standard",
            filename="standard.nuitka-package.config.yml",
            assume_yes_for_downloads=assume_yes_for_downloads,
            check_checksums=check_checksums,
        )
        _package_config.update(
            parsePackageYaml(
                logger=logger,
                package_name="nuitka.plugins.standard",
                filename="stdlib2.nuitka-package.config.yml",
                assume_yes_for_downloads=assume_yes_for_downloads,
                check_checksums=check_checksums,
            )
        )
        _package_config.update(
            parsePackageYaml(
                logger=logger,
                package_name="nuitka.plugins.standard",
                filename="stdlib3.nuitka-package.config.yml",
                assume_yes_for_downloads=assume_yes_for_downloads,
                check_checksums=check_checksums,
            )
        )

        try:
            _package_config.update(
                parsePackageYaml(
                    logger=logger,
                    package_name="nuitka.plugins.commercial",
                    filename="commercial.nuitka-package.config.yml",
                    assume_yes_for_downloads=assume_yes_for_downloads,
                    check_checksums=check_checksums,
                )
            )
        except IOError:
            # No commercial configuration found.
            pass

        # User or plugin provided filenames, but we want PRs though, and will nag
        # about it somewhat.
        for user_yaml_filename in getUserProvidedYamlFiles():
            _package_config.update(
                PackageConfigYaml(
                    logger=logger,
                    name=user_yaml_filename,
                    file_data=getFileContents(user_yaml_filename, mode="rb"),
                    assume_yes_for_downloads=assume_yes_for_downloads,
                    check_checksums=check_checksums,
                )
            )

    return _package_config


def getYamlPackageConfigurationSchemaFilename():
    """Get the filename of the schema for Nuitka package configuration."""
    return os.path.join(
        os.path.dirname(__file__),
        "..",
        "package_config",
        "nuitka-package-config-schema.json",
    )


#     Part of "Nuitka", an optimizing Python compiler that is compatible and
#     integrates with CPython, but also works on its own.
#
#     Licensed under the GNU Affero General Public License, Version 3 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.gnu.org/licenses/agpl.txt
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.



#     Copyright 2025, Kay Hayen, mailto:kay.hayen@gmail.com find license text at end of file


"""Nested loop variable type inference and post-loop operations."""

from __future__ import print_function

# nuitka-project: --python-flag=no_warnings


# for+for: None -> list, iterate after nested loops


def functionForForListIterate():
    data = [[1, 2, 3]]
    collected = None
    for group in data:
        for item in group:
            if collected is None:
                collected = []
            collected.append(item)
    out = []
    if collected is not None:
        for x in collected:
            out.append(x)
    print("for+for list iterate:", out)


# for+while: None -> list, iterate after nested loops


def functionForWhileListIterate():
    data = [[3, 2, 1]]
    collected = None
    for group in data:
        while group:
            item = group.pop()
            if collected is None:
                collected = []
            collected.append(item)
    out = []
    if collected is not None:
        for x in collected:
            out.append(x)
    print("for+while list iterate:", sorted(out))


#  while+for: None -> list, iterate after nested loops


def functionWhileForListIterate():
    data = [[1, 2, 3]]
    collected = None
    idx = 0
    while idx < len(data):
        for item in data[idx]:
            if collected is None:
                collected = []
            collected.append(item)
        idx += 1
    out = []
    if collected is not None:
        for x in collected:
            out.append(x)
    print("while+for list iterate:", out)


# for+for: None -> dict, iterate keys after nested loops


def functionForForDictIterate():
    data = [["a", "b"], ["c"]]
    mapping = None
    for group in data:
        for ch in group:
            if mapping is None:
                mapping = {}
            mapping[ch] = ord(ch)
    result = []
    if mapping is not None:
        for k in mapping:
            result.append(k)
    print("for+for dict iterate:", sorted(result))


# for+for: None -> set, iterate members after nested loops


def functionForForSetIterate():
    data = [[1, 2, 2], [3, 1]]
    seen = None
    for group in data:
        for item in group:
            if seen is None:
                seen = set()
            seen.add(item)
    result = []
    if seen is not None:
        for x in seen:
            result.append(x)
    print("for+for set iterate:", sorted(result))


# Closure with nested loops and callback iteration


def functionClosureCallbackIterate():
    output = []
    tasks = [[lambda: output.append("a"), lambda: output.append("b")]]

    def run():
        ready = None
        for batch in tasks:
            while batch:
                fn = batch.pop()
                if ready is None:
                    ready = []
                ready.append(fn)
        if ready is not None:
            for f in ready:
                f()

    run()
    print("closure callback iterate:", sorted(output))


# Single loop (no nesting) with iteration - control case


def functionSingleLoopIterate():
    data = [1, 2, 3]
    collected = None
    for item in data:
        if collected is None:
            collected = []
        collected.append(item)
    out = []
    if collected is not None:
        for x in collected:
            out.append(x)
    print("single loop iterate:", out)


# Chained inner loops (canary): inner loop 1 builds list, inner loop 2
# iterates it, within the same outer body.


def functionChainInnerLoops():
    groups = [[1, 2], [3, 4]]
    total = 0
    for group in groups:
        bucket = None
        for val in group:
            if bucket is None:
                bucket = []
            bucket.append(val * 10)
        for item in bucket:
            total += item
    print("chain inner loops:", total)


# Variable used between outer iterations (canary): shape must propagate
# to outer loop body after inner loop completes.


def functionUseBetweenOuterIters():
    groups = [[1, 2], [3, 4], [5]]
    result = None
    for group in groups:
        for val in group:
            if result is None:
                result = []
            result.append(val)
        result.append(0)
    print("between outer iters:", result)


# ---------------------------------------------------------------------------
# 1. Attribute / method calls (canary)
# ---------------------------------------------------------------------------


def functionMethodAppend():
    """list.append() on variable built in nested loop."""
    data = [[1, 2], [3]]
    result = None
    for group in data:
        for val in group:
            if result is None:
                result = []
            result.append(val)
    result.append(99)
    print("method_append:", result)


# ---------------------------------------------------------------------------
# 2. Subscript / indexing (canary)
# ---------------------------------------------------------------------------


def functionListIndexing():
    """Indexing a list built in nested loop."""
    data = [[10, 20], [30]]
    result = None
    for group in data:
        for val in group:
            if result is None:
                result = []
            result.append(val)
    print("indexing:", result[0], result[-1])


# ---------------------------------------------------------------------------
# 3. Binary / augmented operations
#    Wrong type -> wrong operator specialization or TypeError.
# ---------------------------------------------------------------------------


def functionListConcatenation():
    """List + list where left operand is from nested loop."""
    data = [[1, 2], [3]]
    result = None
    for group in data:
        for val in group:
            if result is None:
                result = []
            result.append(val)
    combined = result + [4, 5]
    print("list_concat:", combined)


def functionListMultiply():
    """List * int where list is from nested loop."""
    data = [[1], [2]]
    result = None
    for group in data:
        for val in group:
            if result is None:
                result = []
            result.append(val)
    doubled = result * 2
    print("list_mult:", doubled)


def functionListIadd():
    """List += on variable from nested loop."""
    data = [[1, 2], [3]]
    result = None
    for group in data:
        for val in group:
            if result is None:
                result = []
            result.append(val)
    result += [4]
    print("list_iadd:", result)


def functionIntArithmetic():
    """Arithmetic on int accumulated in nested loop."""
    data = [[1, 2], [3, 4]]
    total = None
    for group in data:
        for val in group:
            if total is None:
                total = 0
            total = total + val
    result = total * 2 + 1
    print("int_arith:", result)


def functionFloatArithmetic():
    """Float arithmetic on value from nested loop."""
    data = [[1.5, 2.5], [3.0]]
    total = None
    for group in data:
        for val in group:
            if total is None:
                total = 0.0
            total = total + val
    result = total / 2.0
    print("float_arith:", result)


def functionStrConcatenation():
    """String + string where left operand is from nested loop."""
    data = [["ab", "cd"], ["ef"]]
    text = None
    for group in data:
        for w in group:
            if text is None:
                text = ""
            text = text + w
    full = text + "!"
    print("str_concat:", full)


def functionStrMultiply():
    """String * int where string is from nested loop."""
    data = [["a"], ["b"]]
    text = None
    for group in data:
        for w in group:
            if text is None:
                text = ""
            text = text + w
    print("str_mult:", text * 3)


# ---------------------------------------------------------------------------
# 4. Boolean context / truthiness (canary)
# ---------------------------------------------------------------------------


def functionTruthinessIf():
    """if <var>: where var is built in nested loop."""
    data = [[1, 2], [3]]
    result = None
    for group in data:
        for val in group:
            if result is None:
                result = []
            result.append(val)
    if result:
        print("truthiness_if: non-empty", len(result))
    else:
        print("truthiness_if: empty")


# ---------------------------------------------------------------------------
# 5. Membership testing (canary)
# ---------------------------------------------------------------------------


def functionInList():
    """'val in list' where list is from nested loop."""
    data = [[1, 2], [3, 4]]
    items = None
    for group in data:
        for val in group:
            if items is None:
                items = []
            items.append(val)
    print("in_list:", 3 in items, 99 in items)


# ---------------------------------------------------------------------------
# 6. Unpacking
#    Optimizer must know the variable is iterable to allow unpacking.
# ---------------------------------------------------------------------------


def functionUnpackList():
    """Unpack a 2-element list built in nested loop."""
    data = [[1], [2]]
    result = None
    for group in data:
        for val in group:
            if result is None:
                result = []
            result.append(val)
    a, b = result
    print("unpack_list:", a, b)


def functionUnpackInLoop():
    """Unpack variable from nested loop inside a for loop."""
    data = [[(1, "a")], [(2, "b")]]
    pairs = None
    for group in data:
        for val in group:
            if pairs is None:
                pairs = []
            pairs.append(val)
    nums = []
    for n, c in pairs:
        nums.append(n)
    print("unpack_in_loop:", nums)


# ---------------------------------------------------------------------------
# 7. len() and builtins (canary)
# ---------------------------------------------------------------------------


def functionLenList():
    """len() on list from nested loop."""
    data = [[1, 2], [3, 4, 5]]
    result = None
    for group in data:
        for val in group:
            if result is None:
                result = []
            result.append(val)
    print("len_list:", len(result))


# ---------------------------------------------------------------------------
# 8. Comparison specialization
# ---------------------------------------------------------------------------


def functionEqComparison():
    """== comparison on variable from nested loop."""
    data = [[1, 2], [3]]
    result = None
    for group in data:
        for val in group:
            if result is None:
                result = []
            result.append(val)
    print("eq:", result == [1, 2, 3], result == [])


def functionListOrdering():
    """< > comparisons on lists from nested loop."""
    data1 = [[1, 2], [3]]
    data2 = [[4, 5], [6]]
    r1 = None
    r2 = None
    for group in data1:
        for val in group:
            if r1 is None:
                r1 = []
            r1.append(val)
    for group in data2:
        for val in group:
            if r2 is None:
                r2 = []
            r2.append(val)
    print("ordering:", r1 < r2, r1 > r2)


# ---------------------------------------------------------------------------
# 9. Iteration protocol (canary)
# ---------------------------------------------------------------------------


def functionIterNext():
    """iter() + next() on variable from nested loop."""
    data = [[10, 20], [30]]
    result = None
    for group in data:
        for val in group:
            if result is None:
                result = []
            result.append(val)
    it = iter(result)
    print("iter_next:", next(it), next(it))


# ---------------------------------------------------------------------------
# 10. Comprehension iteration
# ---------------------------------------------------------------------------


def functionListCompOnLoopVar():
    """List comprehension over variable from nested loop."""
    data = [[1, 2], [3]]
    result = None
    for group in data:
        for val in group:
            if result is None:
                result = []
            result.append(val)
    doubled = [x * 2 for x in result]
    print("list_comp:", doubled)


def functionSetCompOnLoopVar():
    """Set comprehension over variable from nested loop."""
    data = [[1, 2, 2], [3, 1]]
    result = None
    for group in data:
        for val in group:
            if result is None:
                result = []
            result.append(val)
    uniq = {x for x in result}
    print("set_comp:", sorted(uniq))


# Execute all tests.
functionForForListIterate()
functionForWhileListIterate()
functionWhileForListIterate()
functionForForDictIterate()
functionForForSetIterate()
functionClosureCallbackIterate()
functionSingleLoopIterate()
functionChainInnerLoops()
functionUseBetweenOuterIters()

functionMethodAppend()
functionListIndexing()
functionListConcatenation()
functionListMultiply()
functionListIadd()
functionIntArithmetic()
functionFloatArithmetic()
functionStrConcatenation()
functionStrMultiply()
functionTruthinessIf()
functionInList()
functionUnpackList()
functionUnpackInLoop()
functionLenList()
functionEqComparison()
functionListOrdering()
functionIterNext()
functionListCompOnLoopVar()
functionSetCompOnLoopVar()

#     Python tests originally created or extracted from other peoples work. The
#     parts were too small to be protected.
#
#     Licensed under the Apache License, Version 2.0 (the "License");
#     you may not use this file except in compliance with the License.
#     You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#     Unless required by applicable law or agreed to in writing, software
#     distributed under the License is distributed on an "AS IS" BASIS,
#     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#     See the License for the specific language governing permissions and
#     limitations under the License.
